module wei_D
  use wei_util 
  use wei_down
  use wei_run
  use wei_external
  use wei_type
  !---------------
  implicit none
contains
  !--------------------------------------------------------------
  !
  !  Function to initialize a data object either from
  ! 
  !    a file or from then general parameters
  !
  !
  !
  logical function data_init(X,tag)
    type(dat),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    data_init=F

    X%tag = tag
    if(isfile(X%tag)) then
       if(.not. data_history_init(X)) then
         write(0,*) "trouble initializing data_history"
         return
       end if
    else
       if(.not. data_param_init(X)) then
         write(0,*) "trouble initializing data_parameter"
         return
       end if
    end if

    data_init=T

  end function data_init
  !--------------------------------------------------------------
  !
  !  Function to initialize source object from input or 
  !          construct a wavelet
  ! 
  logical function source_init(X,tag)
    type(dat),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    source_init=F

    X%tag = tag
    if(isfile(X%tag)) then
       if(.not. source_history_init(X) ) then
         write(0,*) "trouble initializing data_history"
         return
       end if
    else
!       if(.not. source_param_init(X)) then
!         write(0,*) "trouble making source function"
!         return
!       end if
       call erexit("Source module not built")
    end if

    source_init=T

  end function source_init

  !----------------------------------------------------------------
  !
  !
  !  Create an auxilary data file
  !
  logical function aux_data_init(X,tag)
    type(dat),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    aux_data_init=F

    X%tag = tag
    if(.not. data_scratch_init(X)) then
       write(0,*) "trouble initializing scratch data file"
       return
    end if

    aux_data_init=T

  end function aux_data_init
  !----------------------------------------------------------------
  !
  !  Initialize a tag from history
  !   check to make sure that is dimensions dont disagree
  !    with the general params
  ! 
  logical function data_history_init(X)
    type(dat), intent(inout) :: X
    integer:: bogus
#ifdef DEBUG
    character(len=128), parameter :: unit='data_history_init'
    if(run%debug) call in(unit)
#endif

    call auxinout(X%tag)
    data_history_init=F

    X%from = "history"
    call from_aux(X%tag,'esize',X%esize,4)
    if(X%esize==4) call seperr("data has wrong esize, input frequency")
!! . . Initialize Axes calls from history file
    call hinaxis(X%tag,X%amx,1); X%amx%l='mx'
    call hinaxis(X%tag,X%amy,2); X%amy%l='my'
    call hinaxis(X%tag,X%aw ,3); X%aw%l='w_'
    call hinaxis(X%tag,X%asx,4); X%asx%l='sx'
    call hinaxis(X%tag,X%asy,5); X%asy%l ='sy'

!! . . geometry check
    if((X%amx%d .lt. 0) .or. (X%amy%d .lt. 0))&
    call seperr("Please refrain from negative d1 and/or d2")
!! . . Default indices to read into prop. grid
    X%indx=1;                      X%indy=1
    X%finx=X%indx + X%amx%n -1;    X%finy=X%indy + X%amy%n -1

    call global_var_calc(X)

    if(.not. data_setup(X,.false.)) then
       write(0,*) "trouble setting up data"
       return
    end if

    if(run%verb) call data_report(X)
    data_history_init=T
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_history_init
  !----------------------------------------------------------------
  !
  ! almost the same as data_history_init, leaving out assigns to 
  !     down structure
  logical function source_history_init(X)
    type(dat), intent(inout) :: X
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='data_source_init'
    if(run%debug) call in(unit)
#endif
    source_history_init=F
    call auxinout(X%tag)
    X%from = "history"
    call from_aux(X%tag,'esize',X%esize,4)
    if(X%esize==4) call seperr("source has wrong esize, input frequency")
!! . . Initialize Axes calls from history file

    X%amx = down%fatx!    call hinaxis(X%tag,X%amx,1)
    X%amy = down%faty!    call hinaxis(X%tag,X%amy,2)
    call hinaxis(X%tag,X%aw ,3)

    X%asx%n=1;X%asx%d=1.;X%asx%o=0. ! need default values
    X%asy%n=1;X%asy%d=1.;X%asy%o=0. ! need default values
!! Frequency comparisson to data file
    if (X%aw%n /= down%aw%n) call erexit("Source file wrong length")   
    if (X%aw%d /= down%aw%d) call erexit("Source file freq. increment wrong")
    if (X%aw%o /= down%aw%o) call erexit("Source file freq. origin wrong")

!! . . Default indices to read into prop. grid
    X%indx=1;                      X%indy=1
    X%finx=X%indx + X%amx%n -1;    X%finy=X%indy + X%amy%n -1

    if(.not. data_setup(X,.false.)) then
       write(0,*) "trouble setting up source files"
       return
    end if

    !!re-allocate array reallocte array after distribution to size of
    !!propagation grid
    deallocate(X%bin)
    allocate(X%bin(down%amx%n,down%amy%n,down%nws),stat=er)
    if(er/=0) call seperr('trouble at source_history_init reallocate D_d.bin')
    X%bin=(0.0,0.0)

    if(down%dcnv) then
       allocate(X%ddbin(size(X%bin,1),size(X%bin,2),size(X%bin,3)),stat=er)
       if(er/=0) call seperr('trouble at source_history_init allocate D_d.ddbin')
       X%ddbin=(0.0,0.0)
    end if
write(0,*) 'D  down%amx%n',down%amx%n
    if(run%verb) call data_report(X)
    source_history_init=T
#ifdef DEBUG
!    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function source_history_init
!----------------------------------------------------------------
subroutine global_var_calc(X)
type(dat),intent(in) :: X
 integer::bogus
 real::n,m
#ifdef DEBUG
    character(len=128), parameter :: unit='global_var_calc'
    if(run%debug) call in(unit)
#endif
!! . . Define propagation wavefield on model surface
 down%amx=X%amx
 down%amy=X%amy

 down%asx%n=max(X%asx%n,down%nx_shots)
 down%asx%o=min(X%asx%o,down%ox_shots)
 down%asx%d=X%asx%d
 down%asy%n=max(X%asy%n,down%ny_shots)
 down%asy%o=min(X%asy%o,down%oy_shots)
 down%asy%d=X%asy%d

if(.not.down%passive) then
   if(down%marine )then!!r is relative to s position. s-pos=actual earth location
      if(X%amx%n > 1)then
         !!working space
                        !-offsets      , +offs
         down%amx%o=min(X%asx%o+X%amx%o, X%asx%o)
           !   -offs  ,+offsets
         n=max(X%asx%o,X%asx%o+X%amx%o+(X%amx%n-1)*X%amx%d) !!last trace pos.
         down%amx%n=1.5+(n-down%amx%o)/down%amx%d !!size working space
         if(.not.down%add) then
            !! total image space
            down%aix%d=down%amx%d
            down%aix%o=min(down%amx%o,down%asx%o+X%amx%o,down%asx%o)
            n=down%asx%o+down%asx%d*(down%asx%n-1)  !!position of last shot
            m=n+X%amx%o+(X%amx%n-1)*X%amx%d!!position of last receiver
            n=max(n,m)!! last image location needed
            down%aix%n=1.5+(n-down%aix%o)/down%aix%d
         end if
      else
         down%aix%n=1;down%aix%d=1.;down%aix%o=0.
      end if
          
      if(X%amy%n > 1)then
         !!working space
                        !-offsets      , +offs
         down%amy%o=min(X%asy%o+X%amy%o, X%asy%o)
           !   -offs  ,+offsets
         n=max(X%asx%o,X%asy%o+X%amy%o+(X%amy%n-1)*X%amy%d)!!last trace pos.
         down%amy%n=1.5+(n-down%amy%o)/down%amy%d !!size working space
         if(.not.down%add) then
            !! total image space
            down%aiy%d=down%amy%d
            n=down%asy%o+down%asy%d*(down%asy%n-1)!! position last shot
            m=n+X%amy%o+(X%amy%n-1)*X%amy%d!!position of last receiver
            n=max(n,m)!! last image location needed
            down%aiy%n=1.5+(n-down%aiy%o)/down%aiy%d
         end if
      else
         down%aiy%n=1;down%aiy%d=1.;down%aiy%o=0.
      end if
   else!!land geometry: actuall positions for both s&r
          !if(run%verb)write(0,*) 'land'
      if(X%amx%n > 1)then
         down%amx%o=min(X%amx%o,X%asx%o)!!
         n=X%asx%o+(X%asx%n-1)*X%asx%d
         m=X%amx%o+(X%amx%n-1)*X%amx%d
         n=max(n,m)
         down%amx%n=1.5+(n-down%amx%o)/down%amx%d
         if(.not.down%add) down%aix=down%amx
      end if
      if(X%amy%n > 1)then
         down%amy%o=min(X%amy%o,X%asy%o)
         n=X%asy%o+(X%asy%n-1)*X%asy%d
         m=X%amy%o+(X%amy%n-1)*X%amy%d
         n=max(n,m)
         down%amy%n=1.5+(n-down%amy%o)/down%amy%d
         if(.not.down%add) down%aiy=down%amy
      else

      end if
   end if
else
   down%aix=down%amx
   down%aiy=down%amy
end if

down%aw=X%aw;    down%bw=down%aw

!         down%amx%n=max(abs(X%amx%o)                      /X%amx%d +1.5,&
!                           (X%amx%o + (X%amx%n-1)*X%amx%d)/X%amx%d +1.5,&
!                                       real(X%amx%n))
!         down%amx%o=min(X%asx%o+X%amx%o, X%asx%o)
!      end if
!      if(X%amy%n > 1)then
!         down%amy%n=max(abs(X%amy%o)                      /X%amy%d +1.5,&
!                           (X%amy%o + (X%amy%n-1)*X%amy%d)/X%amy%d +1.5,&
!                                       real(X%amy%n))
!         down%amy%o=min(X%asy%o+X%amy%o, X%asy%o)
!      end if
!if(run%verb)write(0,*) 'marine global_var_calc'
!if(run%verb)write(0,*) 'vars'
!if(run%verb)write(0,*) abs(X%amx%o),X%amx%d,X%amx%o,X%amx%n-1,(X%amx%n-1)*X%amx%d,down%fatx%n
!if(run%verb)write(0,*) 'equns'
!if(run%verb)write(0,*) abs(X%amx%o)/X%amx%d +1.5,(X%amx%o + (X%amx%n-1)*X%amx%d)/X%amx%d +1.5
!if(run%verb)write(0,*) 'reulst'
!if(run%verb)write(0,*) max(abs(X%amx%o)/X%amx%d +1.5,(X%amx%o + (X%amx%n-1)*X%amx%d)/X%amx%d +1.5),&
!    max(abs(X%amx%o)/X%amx%d +1.5,(X%amx%o + (X%amx%n-1)*X%amx%d)/X%amx%d +1.5)+(down%fatx%n-1)/2.
!if(run%verb)write(0,*) 'on to the origin'
!if(run%verb)write(0,*) 'eqns'
!if(run%verb)write(0,*) X%asx%o+X%amx%o,X%asx%o - (down%fatx%n-1)*down%amx%d/2.,min(X%asx%o+X%amx%o,X%asx%o - (down%fatx%n-1)*down%amx%d/2.)
!if(run%verb)write(0,*) 'result'
!if(run%verb)write(0,*) down%amx%o
!   else
!      if(X%amx%n > 1)then
!         down%amx%o=min(X%amx%o, X%asx%o - (down%fatx%n-1)*down%amx%d/2.)
!         down%amx%n=down%amx%n+(down%faty%n-1)
!!if(run%verb)write(0,*) 'land global_var_calc'
!!if(run%verb)write(0,*) 'down_amx_n,o',down%amx%n,down%amx%o
!      end if
!      if(X%amy%n > 1)then
!         down%amy%o=min(X%amy%o, X%asy%o - (down%faty%n-1)*down%amy%d/2.)
!         down%amy%n=down%amy%n+(down%faty%n-1)
!if(run%verb)write(0,*) 'land global_var_calc',X%tag
!if(run%verb)write(0,*) 'down down%amy%o, %n',down%amy%o, down%amy%n
!if(run%verb)write(0,*) 'down X%amy%o, %n',X%amy%o, X%amy%n
!      end if
!   end if
!end if
!! . . Set frequency values
!    down%aw=X%aw;    down%bw=down%aw
!   
!   if((.not.down%passive) .and. (down%marine) ) then 
!!! . . image space setup x
!      if(X%amx%n .gt.1) then
!      if(X%amx%o < 0) then
!         down%aix%n=(1.5+(abs(X%amx%o)                      + (X%asx%n-1)*X%asx%d)/X%amx%d) !!no +offs
!              bogus=(1.5+    (          (X%amx%n-1)*X%amx%d + (X%asx%n-1)*X%asx%d)/X%amx%d) !!- and +offs
!         down%aix%n=max(down%aix%n,bogus)
!         down%aix%o=X%asx%o + X%amx%o
!      else
!         down%aix%n=(1.5+    (X%amx%o + (X%amx%n-1)*X%amx%d + (X%asx%n-1)*X%asx%d)/X%amx%d) !!no -offs
!         down%aix%o=X%asx%o
!      end if
!      end if
!
!!! . . image space setup y
!      if(X%amy%n .gt.1) then
!      if(X%amy%o < 0) then
!         down%aiy%n=(1.5+(abs(X%amy%o)                      + (X%asy%n-1)*X%asy%d)/X%amy%d) !!no +offs
!              bogus=(1.5+    (          (X%amy%n-1)*X%amy%d + (X%asy%n-1)*X%asy%d)/X%amy%d) !!- and +offs
!         down%aiy%n=max(down%aiy%n,bogus)
!         down%aiy%o=X%asy%o + X%amy%o
!      else
!         down%aiy%n=(1.5+    (X%amy%o + (X%amy%n-1)*X%amy%d + (X%asy%n-1)*X%asy%d)/X%amy%d)!!no -offs
!         down%aiy%o=X%asy%o
!      end if
!      end if
!
!   end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
end subroutine global_var_calc
!----------------------------------------------------------------
!      
!   relocate wavefield within the model
!    remember: amx%o is a relative distance from the shot!
  subroutine data_position(X,isx,isy)
    type(dat),intent(in) :: X
    integer :: isx,isy

       down%amx%o= min( X%amx%o + X%asx%o+X%asx%d*(isx-1), X%asx%o+X%asx%d*(isx-1) ) 
       down%amy%o= min( X%amy%o + X%asy%o+X%asy%d*(isy-1), X%asy%o+X%asy%d*(isy-1) ) 
     end subroutine data_position
!----------------------------------------------------------------
  subroutine data_position_report()
    if (run%verb)write(0,'(a,f9.3)')"         *  Working grid positioned at x=",&
                                     down%amx%o
    if (run%verb)write(0,'(a,f9.3)')"         *                             y=",&
                                     down%amy%o
  end subroutine data_position_report
!----------------------------------------------------------------
  !
  !
  !
  logical function create_data_file(X,tag)
    type(dat), intent(in) :: X
    character(len=*)         :: tag
       create_data_file=F 
       call auxscr(tag)
       call to_history("esize",8,tag)
       call houaxes(tag,X%amx,X%amy,X%aw,X%asx,X%asy)
       create_data_file=T 
   end function 
  !----------------------------------------------------------------
  logical function data_param_init(X)
    type(dat), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='data_param_init' 
    if(run%debug) call in(unit)
#endif
    data_param_init=F

    X%from = "parameter"
    X%amx = down%amx; X%amy = down%amy; 
    X%asx = down%asx; X%asy = down%asy; 
    X%aw  = down%aw;  
    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble creating data file"
      return
    end if
    if(.not. data_setup(X,.true.)) then
      write(0,*) "trouble setting up data"
      return
    end if

    data_param_init=T
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_param_init

!----------------------------------------------------------------
  subroutine data_assign(X,val)
   type(dat) :: X
   complex :: val
   logical ::a

    a=file_val_assign(X%tag,val)
  end subroutine data_assign
  !----------------------------------------------------------------
  !
  ! data_scratch_init
  !
  !  Create a data scratch file
  !
  !
  logical function data_scratch_init(X)
    type(dat), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='data_scratch_init' 
    if(run%debug) call in(unit)
#endif

    data_scratch_init=F
    X%from = "parameter"
    X%amx = down%amx; X%amy = down%amy; 
    X%asx = down%asx; X%asy = down%asy
    X%aw  = down%aw; 
    if (down%teleseismic) then
       X%asx%d=0.  ; X%asy%d = 0.
    endif
    X%indx=1;                      X%indy=1
    X%finx=X%indx + X%amx%n -1;    X%finy=X%indy + X%amy%n -1

    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble  creating data file from scratch"
      return
    end if
    if(.not. data_setup(X,.true.)) then
      write(0,*) "trouble setting up data from scratch"
      return
    end if
    data_scratch_init=T
    if(run%verb)call data_report(X)
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_scratch_init
!----------------------------------------------------------------  
  subroutine data_passive_pointer(X,Y)
    type (dat),intent(out):: X 
    type (dat),intent(in) :: Y


    down%D_d = down%D; 
    X%bites= Y%bites
    X%tag  = Y%tag; X%tag_0   = Y%tag_0
    X%amx  = Y%amx; X%amy     = Y%amy  
    X%asx  = Y%asx; X%asy     = Y%asy
    X%aw   = Y%aw ; X%aw_local= Y%aw_local
    X%bin  =>Y%bin
  end subroutine data_passive_pointer
  !----------------------------------------------------------------
  ! 
  !  The core  routine to setup data
  logical function data_setup(X,zero)
    type(dat), intent(inout) :: X
    logical :: zero
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='data_setup' 
    if(run%debug) call in(unit)
#endif
    data_setup=F
    X%bites = 8* product( (/X%amx%n,X%amy%n,down%nws/) )

    !! FREQUENCY blocking
    if( (1.0*X%aw_local%n)/down%nws /= X%aw_local%n/down%nws ) then
       call data_report(X)
       call seperr('D.aw.n is not a multiple of nws')
    end if
    down%nwb = down%nw_local 

    !allocate array
!      allocate(X%bin(down%amx%n,down%amy%n,down%nws),stat=er)
      allocate(X%bin(X%amx%n,X%amy%n,down%nws),stat=er)
      if(er/=0) call seperr('cant allocate D.bin')
      X%bin=(0.0,0.0)

    if(zero .and. run%m_nth==1) then
      if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
        write(0,*) "trouble setting ",trim(X%tag_0)
        return
      end if
    end if
   if(run%mpi) then
     if(.not.  mpi_data_setup(X,zero)) then
       write(0,*) "trouble initializng data mpi"
       return
     end if
   else
     X%aw_local=X%aw
   end if

   data_setup=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_setup

  !----------------------------------------------------------------
  !
  !   Setup data structure for local files if
  !     we are running MPI

  logical function mpi_data_setup(X,zero)
    type(dat), intent(inout) :: X
    integer ::n
    logical :: zero
    mpi_data_setup=F

    X%tag_0=X%tag
    write(X%tag,"(a,a)") trim(X%tag_0),run%tag_num

    X%aw_local%n=down%nw_local*down%nws

    n=X%aw%n
    X%aw%n=down%nw_local*down%nws
    
    if(.not. create_data_file(X,X%tag)) then
       write(0,*) "trouble setting up mpi data"
       return
    end if
    X%aw%n=n
       
    if(.not. zero) then 
       call distribute_data(X)
    else
       !       if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
       !          write(0,*) "trouble setting ",trim(X%tag_0)
       !          return
       !       end if
    end if
       
    mpi_data_setup=T
  end function mpi_data_setup
!----------------------------------------------------------------
! Information on the data
!
  subroutine data_report(X)
    type(dat), intent(in) :: X
    write(0,*)'        ************************************************'
    write(0,*)'        *   ',trim(X%tag),' from ',trim(X%from)
    write(0,*)                         "        *              [n]     [o]      [d] "       
    write(0,'(a,i3,f9.2,f9.2)       ') "         *        Offx: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i3,f9.2,f9.2)       ') "         *        Offy: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i3,f9.2,f9.2)       ') "         *        Shtx: ",X%asx%n, X%asx%o, X%asx%d
    write(0,'(a,i3,f9.2,f9.2)       ') "         *        Shty: ",X%asy%n, X%asy%o, X%asy%d
    write(0,'(a,i4,a,f7.3,f9.3,a,i3,a)') "         *      localW:(",X%aw_local%n,")",X%aw%o,X%aw%d,&
    " =",down%nwb," blocks"
    if(associated(X%bin)) then
       write(0,'(a,f9.2,a,f9.2,a)        ') "         *     block: ",X%bites/1024./1024.," / ",8*size(X%bin)/1024./1024.," Mbytes"
    end if
  end subroutine data_report
!--------------------------------------------------------------
! Read the data 
  logical function data_read(X,k)
    type(dat), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat,i,j,l
#ifdef DEBUG
    character(len=128), parameter :: unit='data_read'
    if(run%debug) call in(unit)
#endif

    data_read=F
    X%bin=(0.0,0.0)

    if(X%bites/= sreed(X%tag, X%bin(X%indx:X%finx,X%indy:X%finy,:), X%bites)) then
       write(0,*) "Badness for file ",trim(X%tag)
       write(0,*) "trouble reading block",k,'machine',run%m_ith  
    else
      data_read=T
    end if

    if(down%scale.ne.1.0) then
!$OMP PARALLEL DO PRIVATE(i,j,k)
       do i=1,size(X%bin,3)
       do j=1,size(X%bin,2)
       do l=1,size(X%bin,1)
          X%bin(l,j,i)=down%scale*X%bin(l,j,i)
       end do
       end do
       end do
!$OMP END PARALLEL DO 
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_read
  !--------------------------------------------------------------
  ! figure out where the source goes in the work space
  subroutine data_indices_land(X,isx,isy)
    type(dat) ::X
    integer::isx,isy
#ifdef DEBUG
    character(len=128), parameter :: unit='calc_indices'
    if(run%debug) call in(unit)
#endif 
    X%indx= 1; X%indy= 1;
if(.not.down%passive .or. .not.down%teleseismic) then
   if(X%amx%n .lt. down%amx%n/2 +1 )then
      X%indx= (1.5+ ((down%asx%o + (isx-1)*down%asx%d) - down%amx%o)/down%amx%d) -(X%amx%n-1)/2
   end if
   if(X%amy%n .lt. down%amy%n/2 +1 )then
      X%indy= (1.5+ ((down%asy%o + (isy-1)*down%asy%d) - down%amy%o)/down%amy%d) -(X%amy%n-1)/2
   end if
end if
    X%finx=X%indx + X%amx%n -1
    X%finy=X%indy + X%amy%n -1
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_indices_land
!----------------------------------------------------------------
! figure out where the source goes in the work space
  subroutine data_indices_marine(U,D)
    type(dat) ::U,D!! Upcoming(data),Downgoing(source)
  !! remember: am_%o is a relative distance from the shot!
#ifdef DEBUG
    character(len=128), parameter :: unit='calc_indices_marine'
    if(run%debug) call in(unit)
#endif
    U%indx=1;    D%indx=1;    U%indy=1;    D%indy=1
    if( .not.down%passive) then
       if (U%amx%n .gt. 1 .and. D%amx%n == 1) then
          if (U%amx%o+U%asx%o .gt. down%amx%o)then 
             U%indx= 1.5+    (U%amx%o)/down%amx%d 
          end if
          D%indx= 1.5+ (U%asx%o-down%amx%o)/down%amx%d
       else
          D%indx=1
       end if
       if (U%amy%n .gt. 1 .and. D%amy%n == 1) then
          if (U%amy%o+U%asy%o .gt. down%amy%o)then 
             U%indy= 1.5+    (U%amy%o)/down%amy%d 
          end if
          D%indy= 1.5+ (U%asy%o-down%amy%o)/down%amy%d
       else
          D%indy=1
       end if

    end if

    U%finx=U%indx + U%amx%n -1
    U%finy=U%indy + U%amy%n -1
    D%finx=D%indx + D%amx%n -1
    D%finy=D%indy + D%amy%n -1

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_indices_marine
!----------------------------------------------------------------

!----------------------------------------------------------------
subroutine source_report(X)
type(dat)::X
if (X%amx%n==1) then
   if(run%verb)write(0,'(a,f9.3)') '         *           Source located at x=',&
            down%amx%o + (X%indx-1)*down%amx%d +((X%amx%n-1)/2)*down%amx%d
else
   if(run%verb)write(0,'(a,f9.3)') '         *           Source located at x=',&
            down%amx%o + X%amx%o
end if
if  (X%amy%n==1) then
if(run%verb)write(0,'(a,f9.3)')    '         *                             y=',&
            down%amy%o + (X%indy-1)*down%amy%d +((X%amy%n-1)/2)*down%amy%d
else
   if(run%verb)write(0,'(a,f9.3)') '         *                             y=',&
            down%amy%o + X%amy%o
end if


end subroutine source_report
 !----------------------------------------------------------------
  logical function data_write(X,k)
    type(dat), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='data_write' 
    if(run%debug) call in(unit)
#endif
    data_write=F 

    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%bites/= srite(X%tag, X%bin, X%bites)) then
      write(0,*) "trouble writing data block",k
      return
    end if

!     call to_history("ifreq",k,X%tag)
 
    data_write=T 
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_write
  !----------------------------------------------------------------
  !
  ! 
  !
  !
  subroutine distribute_data(X)
    type(dat), intent(inout) :: X
    integer, external :: sseek
    integer :: ierr,i,j
#ifdef SEP_MPI
    if(run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(X%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)
       do j=1, X%asy%n
          do i=1, X%asx%n
             if(0/=mpi_sep_tag_distribute(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
             down%nwp, down%iw_thread)) &
             call seperr("trouble distributing data ");
          end do
       end do
    end if
    ierr=sseek(trim(X%tag),0,0)
#endif
  end subroutine distribute_data
  !----------------------------------------------------------------
  subroutine combine_data(X)
    type(dat), intent(inout) :: X
    integer :: ierr,i
    integer, external :: sseek
#ifdef SEP_MPI
    i=0;if(down%add) i=1
    if( .not. down%adj .and. run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(X%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)

       if(0/=mpi_sep_tag_combine(run%m_ith,X%tag,0,X%tag_0, &
       X%bites/4,down%nwp,down%iw_thread,i)) &
       call seperr("trouble combining data ");
    end if
    call line()
!    write(0,*) "        *         UNLINKING ",trim(X%tag)
    call aux_unlink(trim(X%tag))
#endif
  end subroutine combine_data

!----------------------------------------------------------------
!  Copy data
  subroutine data_copy_conjg(X,Y)
    type(dat), intent(inout) :: X
    type(dat), intent(inout) :: Y
    integer :: iwb,i,j,k
!    logical :: res
#ifdef DEBUG
    character(len=128), parameter :: unit='data_copy'
    if(run%debug) call in(unit)
#endif
    if(X%bites/=Y%bites) call seperr("D and D_d not allocated same size")
    if (.not.down%fwdscat) then
!$OMP PARALLEL DO PRIVATE(i,j,k) 
       do i=1,size(X%bin,3)
       do j=1,size(X%bin,2)
       do k=1,size(X%bin,1)
          Y%bin(k,j,i) = conjg(X%bin(k,j,i))
       end do; end do; end do
!$OMP END PARALLEL DO 
    else
!$OMP PARALLEL DO PRIVATE(i,j,k) 
       do i=1,size(X%bin,3)
       do j=1,size(X%bin,2)
       do k=1,size(X%bin,1)
          Y%bin(k,j,i) = X%bin(k,j,i)
       end do; end do; end do
!$OMP END PARALLEL DO 
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_copy_conjg
!----------------------------------------------------------------
!  Copy data
  subroutine data_conjg(X)
    type(dat), intent(inout) :: X
    integer :: i,j,k
#ifdef DEBUG
    character(len=128), parameter :: unit='data_conjg'
    if(run%debug) call in(unit)
#endif
    if (down%fwdscat) then
       return
    else
!$OMP PARALLEL DO PRIVATE(i,j,k) 
       do i=1,size(X%bin,3)
       do j=1,size(X%bin,2)
       do k=1,size(X%bin,1)
          X%bin(k,j,i) = conjg(X%bin(k,j,i))
       end do; end do; end do
!$OMP END PARALLEL DO 
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_conjg
!----------------------------------------------------------------
!  Add two data structures
! 
  subroutine data_add(X,Y,rereadX,rereadY)
    type(dat), intent(inout) :: X
    type(dat), intent(inout) :: Y
    integer :: iwb,i,j,k
    logical :: res,rereadX,rereadY
#ifdef DEBUG
    character(len=128), parameter :: unit='data_add'
    if(run%debug) call in(unit)
#endif

    do iwb=1,down%nwb
       if (rereadX) res= data_read (X,iwb)
       if (rereadY) res= data_read (Y,iwb)
!$OMP PARALLEL DO PRIVATE(iwb,i,j,k)
       do i=1,size(X%bin,3)
       do j=1,size(X%bin,2)
       do k=1,size(X%bin,1)
          Y%bin(k,j,i) = Y%bin(k,j,i) + X%bin(k,j,i)
       end do;end do;end do
!$OMP END PARALLEL DO 
       res= data_write(Y,iwb)
    end do
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_add
!----------------------------------------------------------------  
end module wei_D





