module wei_SPig
  !! image up- and down-going wave-fields for Shot Profile migration 
  !!      output is offset domain

  use wei_util
  use wei_external
  use wei_run
  use wei_down
  use wei_flags

  !---------------
  implicit none
  integer ::o_offx,o_offy, zoffx, zoffy
contains
  !----------------------------------------------------------------  
  !! offset domain Shot-profile imaging
  integer function ig_init() result(st)

#ifdef DEBUG
    character(len=128), parameter :: unit='SPimag_init'
    if(run%debug) call in(unit)
#endif
    o_offx=down%ahx%o/down%ahx%d
    o_offy=down%ahy%o/down%ahy%d
    zoffy=1+down%ahy%n/2
    zoffx=1+down%ahx%n/2
    st=0
  end function Ig_init
  !----------------------------------------------------------------
  integer function Higop(model,wfld_u,wfld_d,ddstar) result(st)
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    complex, dimension(:,:,:,:)  , pointer :: model
    real, dimension(:,:,:), pointer :: ddstar
   integer   :: iws,iy,ih
#ifdef DEBUG
    character(len=128), parameter :: unit='SPimag'
    if(run%depar) call in(unit)
!    if(run%verb)write(0,*) 'imwf/model in igop=',cabs(sum(model))
#endif
    if(down%amy%n.gt.1) then
!$OMP PARALLEL DO PRIVATE(iy)
       do iy=1,down%amy%n
          call para_y(iy,model,wfld_u,wfld_d,ddstar)
       end do
!$OMP END PARALLEL DO 
    else
!$OMP PARALLEL DO PRIVATE(ih)
       do ih=1,down%ahx%n
          call para_h(ih,model,wfld_u,wfld_d,ddstar)
       end do
!$OMP END PARALLEL DO 
    end if

!    if (down%image_freqs) call image_freq_write(R,model)

#ifdef DEBUG
    if(run%verb)write(0,*) 'imwf/model after loops igop=',cabs(sum(model))
    if(run%depar) call out(unit)
#endif
    st=0
  end function Higop
!----------------------------------------------------------------
  integer function MultIGop(model,wfld_u,wfld_d,ddstar) result(st)
    complex, dimension(:,:,:), pointer:: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,iy,ih
#ifdef DEBUG
    character(len=128), parameter :: unit='MultIGop'
    if(run%depar) call in(unit)
    if(run%verb)write(0,*) 'imwf/model in Mult-Igop=',cabs(sum(model))
#endif
    if(down%amy%n.gt.1) then
!$OMP PARALLEL DO PRIVATE(iy)
       do iy=1,down%amy%n
          call para_ym(iy,model,wfld_u,wfld_d,ddstar)
       end do
!$OMP END PARALLEL DO 
    else
!$OMP PARALLEL DO PRIVATE(ih)
       do ih=1,down%ahx%n
          call para_hm(ih,model,wfld_u,wfld_d,ddstar)
       end do
!$OMP END PARALLEL DO 
    end if
!    if (down%image_freqs) call image_freq_write(R,model)

#ifdef DEBUG
    if(run%verb)write(0,*) 'imwf/model after loops Mult-Igop=',cabs(sum(model))
    if(run%depar) call out(unit)
#endif
    st=0
  end function MultIGop
!----------------------------------------------------------------
  integer function NdIGop(model,wfld_u,wfld_d,ddstar) result(st)
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,iy,ih
#ifdef DEBUG
    character(len=128), parameter :: unit='Nd-IGop'
    if(run%depar) call in(unit)
    if(run%verb)write(0,*) 'imwf/model in Nd-Igop=',cabs(sum(model))
#endif
    if(down%amy%n.gt.1) then
!$OMP PARALLEL DO PRIVATE(iy)
       do iy=1,down%amy%n
          call para_yn(iy,model,wfld_d)
       end do
!$OMP END PARALLEL DO 
    else
!$OMP PARALLEL DO PRIVATE(ih)
       do ih=1,down%ahx%n
          call para_hn(ih,model,wfld_d)
       end do
!$OMP END PARALLEL DO 
    end if
!    if (down%image_freqs) call image_freq_write(R,model)

#ifdef DEBUG
    if(run%verb)write(0,*) 'imwf/model after loops Nd-Igop=',cabs(sum(model))
    if(run%depar) call out(unit)
#endif
    st=0
  end function NdIGop
!----------------------------------------------------------------
  integer function NuIGop(model,wfld_u,wfld_d,ddstar) result(st)
    complex, dimension(:,:,:), pointer:: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,iy,ih
#ifdef DEBUG
    character(len=128), parameter :: unit='Nu_imag'
    if(run%depar) call in(unit)
    if(run%verb)write(0,*) 'imwf/model in Nu_igop=',cabs(sum(model))
#endif
    if(down%amy%n.gt.1) then
!$OMP PARALLEL DO PRIVATE(iy)
       do iy=1,down%amy%n
          call para_yn(iy,model,wfld_u)
       end do
!$OMP END PARALLEL DO 
    else
!$OMP PARALLEL DO PRIVATE(ih)
       do ih=1,down%ahx%n
          call para_hn(ih,model,wfld_u)
       end do
!$OMP END PARALLEL DO 
    end if
!    if (down%image_freqs) call image_freq_write(R,model)

#ifdef DEBUG
    if(run%verb)write(0,*) 'imwf/model after loops Nu_igop=',cabs(sum(model))
    if(run%depar) call out(unit)
#endif
    st=0
  end function NuIGop
!----------------------------------------------------------------
  subroutine para_h(ih,model,wfld_u,wfld_d,ddstar)
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,ih,it,f1,f2,j1,j2,i1
#ifdef DEBUG
    character(len=128), parameter :: unit='para_h'
    if(run%depar) call in(unit)
#endif
    if(.not.down%dcnv) then
       freq:do iws=1,down%nws
          it=ih+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld_u,1),min(size(wfld_u,1)-f1,size(wfld_u,1)-f2))
!          do i1=j1,j2 ! why didnt this loop work?
!             model(i1,1,ih,zoffy)=model(11,1,ih,zoffy) + &
!                 wfld_u( i1+f2, 1,iws)*wfld_d(i1+f1, 1,iws)
!          end do
          model(j1:j2,1,ih,zoffy)=model(j1:j2,1,ih,zoffy) + &
             wfld_u( j1+f2:j2+f2, 1,iws)*wfld_d(j1+f1:j2+f1, 1,iws)
             !! conjugate taken care of after data_read ()

!call seperr("asdf")
       end do freq
    else
       freqdec:do iws=1,down%nws
          it=ih+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld_u,1),min(size(wfld_u,1)-f1,size(wfld_u,1)-f2))
          !! conjugate taken care of after data_read ()          
          model(j1:j2,1,ih,zoffy)=model(j1:j2,1,ih,zoffy) + &
             wfld_u( j1+f2:j2+f2, 1,iws)*wfld_d(j1+f1:j2+f1, 1,iws)/ &
             ddstar(j1+f2:j2+f2, 1,iws)

       end do freqdec
    end if
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine para_h 
!----------------------------------------------------------------
  subroutine para_y(iy,model,wfld_u,wfld_d,ddstar)
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,iy,i,it,f1,f2,j1,j2
#ifdef DEBUG
    character(len=128), parameter :: unit='para_y'
    if(run%depar) call in(unit)
#endif
    if(.not.down%dcnv) then
    freq:do iws=1,down%nws
       xoffs:do i=1,down%ahx%n
          it=i+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld_u,1),min(size(wfld_u,1)-f1,size(wfld_u,1)-f2))         
          model(j1:j2,iy,i,zoffy)=model(j1:j2,iy,i,zoffy) + &
             wfld_u( j1+f2:j2+f2, iy,iws)!*wfld_d(j1+f1:j2+f1, iy,iws)
            !! conjugate taken care of after data_read ()
       end do xoffs
    end do freq
    else
    freqdec:do iws=1,down%nws
       xoffsd:do i=1,down%ahx%n
          it=i+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld_u,1),min(size(wfld_u,1)-f1,size(wfld_u,1)-f2))
           !! conjugate taken care of after data_read ()          
          model(j1:j2,iy,i,zoffy)=model(j1:j2,iy,i,zoffy) + &
             wfld_u( j1+f2:j2+f2, iy,iws)*wfld_d(j1+f1:j2+f1, iy,iws)/ &
             ddstar(j1+f2:j2+f2, 1,iws)

       end do xoffsd
    end do freqdec
    end if
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine para_y
!----------------------------------------------------------------
  subroutine para_hm(ih,model,wfld_u,wfld_d,ddstar)
    !!SRME multiples version 
    complex, dimension(:,:,:), pointer:: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,ih,it,f1,f2,j1,j2
#ifdef DEBUG
    character(len=128), parameter :: unit='para_hm'
    if(run%depar) call in(unit)
#endif
    if(down%dcnv) then
       freq:do iws=1,down%nws
          it=ih+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld_u,1),min(size(wfld_u,1)-f1,size(wfld_u,1)-f2))         
          model(j1:j2,1,ih,zoffy)=model(j1:j2,1,ih,zoffy) + &
               wfld_u(j1+f1:j2+f1, 1,iws)**2 *conjg(wfld_d( j1+f2:j2+f2, 1,iws))/ &
               ddstar(j1+f2:j2+f2, 1,iws)
!*conjg(wfld_d( j1+f2:j2+f2, 1,iws))  &
!*wfld_d( j1+f2:j2+f2, 1,iws)*conjg(wfld_d( j1+f2:j2+f2, 1,iws))
       end do freq
    else
       freqd:do iws=1,down%nws
          it=ih+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld_u,1),min(size(wfld_u,1)-f1,size(wfld_u,1)-f2))         
          model(j1:j2,1,ih,zoffy)=model(j1:j2,1,ih,zoffy) + &
               wfld_u(j1+f1:j2+f1, 1,iws)*wfld_u( j1+f2:j2+f2, 1,iws)
       end do freqd
    end if
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine para_hm 
!----------------------------------------------------------------
  subroutine para_ym(iy,model,wfld_u,wfld_d,ddstar)
    !!SRME multiples version 
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,iy,i,it,f1,f2,j1,j2
#ifdef DEBUG
    character(len=128), parameter :: unit='para_ym'
    if(run%depar) call in(unit)
#endif

    freq:do iws=1,down%nws
       xoffs:do i=1,down%ahx%n
          it=i+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld_u,1),min(size(wfld_u,1)-f1,size(wfld_u,1)-f2))          
          model(j1:j2,iy,i,zoffy)=model(j1:j2,iy,i,zoffy) + &
             wfld_u( j1+f2:j2+f2, iy,iws)*wfld_u(j1+f1:j2+f1, iy,iws) / &
             ddstar(j1+f2:j2+f2, iy,iws)

       end do xoffs
    end do freq
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine para_ym
!----------------------------------------------------------------
  subroutine para_hn(ih,model,wfld)
    !!SRME multiples version 
    complex, dimension(:,:,:), pointer :: wfld
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,ih,it,f1,f2,j1,j2
#ifdef DEBUG
    character(len=128), parameter :: unit='para_hn'
    if(run%depar) call in(unit)
#endif
    freq:do iws=1,down%nws
       it=ih+o_offx-1
       it=2*it
       f1=-floor((real(it)+0.1)/2.)
       f2=f1+it
       j1=max(1,max(1-f1,1-f2))
       j2=min(size(wfld,1),min(size(wfld,1)-f1,size(wfld,1)-f2))          
       model(j1:j2,1,ih,zoffy)=model(j1:j2,1,ih,zoffy) + &
             cabs(wfld(j1+f1:j2+f1, 1,iws))

    end do freq
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine para_hn 
!----------------------------------------------------------------
  subroutine para_yn(iy,model,wfld)
    !!SRME multiples version 
    complex, dimension(:,:,:), pointer:: wfld
    complex, dimension(:,:,:,:)  , pointer :: model
    integer   :: iws,iy,i,it,f1,f2,j1,j2
#ifdef DEBUG
    character(len=128), parameter :: unit='para_yn'
    if(run%depar) call in(unit)
#endif
    freq:do iws=1,down%nws
       xoffs:do i=1,down%ahx%n
          it=i+o_offx-1
          it=2*it
          f1=-floor((real(it)+0.1)/2.)
          f2=f1+it
          j1=max(1,max(1-f1,1-f2))
          j2=min(size(wfld,1),min(size(wfld,1)-f1,size(wfld,1)-f2))          
          model(j1:j2,iy,i,zoffy)=model(j1:j2,iy,i,zoffy) + &
             cabs(wfld( j1+f2:j2+f2, iy,iws))

       end do xoffs
    end do freq
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine para_yn

!!----------------------------------------------------------------
!  subroutine image_freq_write(R,model)
!    complex, dimension(:,:,:,:)  , pointer :: model
!    type(imag) ::R
!    integer   :: i1,i2,i3,i4
!          do i1=1,size(model,4)
!          do i2=1,size(model,3)
!          do i3=1,size(model,2)
!          do i4=1,size(model,1)
!             R%bin(i1,i2,i3,i4,1)=model(i1,i2,i3,i4)
!          end do
!          end do
!          end do
!          end do
!
!          if (.not. image_insert(R))&
!             call seperr("Single shot image => global image didnt go well")
!          do i4=1,size(R%bin_tot,4)
!          do i3=1,size(R%bin_tot,3)
!             call image_write(izbasdfasdf)
!             call write_mx_my(R%tag,R%esize,size(R%bin_tot,1),size(R%bin_tot,2),&
!                  R%bin(:,:,i3,i4,1),down%shotnum/=0)
!          end do
!          end do
!  end subroutine image_freq_write
!----------------------------------------------------------------
end module wei_SPig
