module wei_R
  use wei_util
  use wei_external
  use wei_run
  use wei_down
  use wei_down_type
  use wei_type
  use wei_lagrange

  implicit none
  integer, private, save,pointer  :: image_master(:)
  integer, private, save  :: image_comm_group
  logical, private, save  :: image_lead_thread
  real,allocatable,dimension(:,:,:,:)::tmpr

contains
!----------------------------------------------------------------
! Image initializing routine
  logical function  image_init(X,tag)
    type(imag),       intent(inout) :: X
    character(len=128), intent(in)    :: tag
    integer ::ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='image_init'
    if(run%debug) call in(unit)
#endif
    image_init=F
    X%tag = tag

    if(down%image_real) then
       allocate(tmpr(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n),stat=ierr)
       if(ierr/=0) call seperr('cant allocate tmpr R.F90',ierr)
    end if

!    call MPI_Barrier(MPI_COMM_WORLD,ierr)
    if(((.not.down%adj) .or. (down%add)) .and. (isfile(X%tag))) then
       if(.not. image_history_init(X)) then
         write(0,*) "trouble initializing image_history"
         return
       end if
    else
       if(.not. image_param_init(X)) then  !THIS ONE
         write(0,*) "trouble initializing image_param"
         return
       end if       
    end if

    if(var%use_disk_buf) then
       write(X%tag_var,"(a,a)") "R_var",run%tag_num
       call auxinout(X%tag_var)
       ierr= srite(X%tag_var,0.,4)
    end if

    image_init=T
#ifdef DEBUG
      if(run%debug) call out(unit)
#endif
  end function image_init
!----------------------------------------------------------------
! Image initializing routine
  logical function  aux_image_init(X,Y,tag)
    type(imag),       intent(out) :: X
    type(imag),       intent(in) :: Y
    character(len=128), intent(in)    :: tag
    integer ::ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='image_init'
    if(run%debug) call in(unit)
#endif
    aux_image_init=F
    X%tag=tag

    if(((.not.down%adj) .or. (down%add)) .and. (isfile(X%tag))) then
       if(.not. image_history_init(X)) then
         write(0,*) "trouble initializing aux_image_history"
         return
       end if
       if(X%amx%n.ne.Y%amx%n.or.X%amy%n.ne.Y%amy%n.or.X%az%n.ne.Y%az%n) &
         call seperr('aux image file not the same size as R')
    else !!for new files    
       X%esize=Y%esize ;X%amx=Y%amx; X%ahx=Y%ahx; X%amy=Y%amy; X%ahy=Y%ahy; X%az=Y%az; X%aw=Y%aw
       X%indx=Y%indx; X%indy=Y%indy; X%finx=Y%finx; X%finy=Y%finy; X%from=Y%from; X%bites=Y%bites
       if(.not. create_image_file(X,X%tag)) then
          write(0,*) "trouble creating image"
          return
       end if
       if(.not. image_setup(X,.not.down%add)) then
          write(0,*) "trouble in image setup"
          return
       end if
    end if
    call image_report(X)
    aux_image_init=T
#ifdef DEBUG
      if(run%debug) call out(unit)
#endif
  end function aux_image_init

!----------------------------------------------------------------
! Initialize an image file based on parameters
  logical function image_param_init(X)
    type(imag), intent(inout) :: X
    integer :: noffx,noffy
#ifdef DEBUG
    character(len=128), parameter :: unit='image_param_init'
    if(run%debug) call in(unit)
#endif
    image_param_init=F
    X%from = "parameter"
    X%esize=4;if(.not.down%image_real) X%esize=8
!! . . depth o-rama
    X%az%d=down%az%d
    X%az%n=down%az%n-down%firstz+1
    X%az%o=down%az%o + down%firstz*down%az%d

!! . . calculate origin for CIGs
    down%ahx%o= -( down%ahx%n -1)/2. * down%amx%d
    down%ahy%o= -( down%ahy%n -1)/2. * down%amy%d !offset origin begins w/ neg. offset
    down%ahx%d=down%amx%d
    down%ahy%d=down%amy%d
    down%ahx%l='hx'
    down%ahy%l='hy'

    X%ahx=down%ahx
    X%ahy=down%ahy
!!total image size from earth covered by recievers
    X%amx=down%aix 
    X%amy%n=1;X%amy%o=0.;X%amy%d=1.
    if (down%amy%n.gt.1)X%amy=down%aiy

    if(down%image_freqs) X%aw=down%aw; X%az%l='z'

    if((run%m_ith==0).and.(.not.down%add) ) then 
      if(.not. create_image_file(X,X%tag)) then
        write(0,*) "trouble creating image"
        return
       end if
    end if

    if(.not. image_setup(X,.not.down%add)) then
      write(0,*) "trouble in image setup"
      return
    end if

    call image_report(X)
    image_param_init=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function image_param_init
!----------------------------------------------------------------
! Write out parameters for an image file
  logical function create_image_file(X,tag)
    type(imag) :: X
    type(dat)::D,D_d
    type(slow)::S
    type(wfld)::W,W_d
    logical :: i_real
    integer :: i,j,k
    character(len=*) ::  tag 
    create_image_file=F

    call auxscr(tag)
    call to_history("esize",X%esize,tag)
    if (down%image_freqs)then
       call houaxes(tag,X%amx,X%amy,X%ahx,X%ahy,X%aw,X%az)
    else
       call houaxes(tag,X%amx,X%amy,X%ahx,X%ahy,X%az)
       call to_history("n6",1 ,tag);call to_history("o6",1 ,tag)
       call to_history("d6",1 ,tag) !!sometimes you want some ones
    end if
    !! General info into
    !call to_history("D" ,trim(D%tag_0) ,tag)
    !call to_history("D_d",trim(D_d%tag_0) ,tag)
    !call to_history("S" ,trim(S%s_s%tag_0) ,tag)
    !if (down%wIOu) call to_history("W" ,trim(W%tag) ,tag)
    !if (down%wIOd) call to_history("W_d" ,trim(W_d%tag) ,tag)
    call to_history("nref",down%nref ,tag)
    call to_history("marine geometry",down%marine,tag)
    call to_history("passive",down%passive ,tag)
    call to_history("forward scattering",down%fwdscat,tag)
    call to_history("velocity",S%s_s%velocity ,tag)
    call to_history("operation",run%operation ,tag)
    call to_history("adj",down%adj ,tag)
    call to_history("add", down%add,tag)
    call to_history("order",down%order ,tag)
    call to_history("converted wave",down%cwave ,tag)
    call to_history("image_freqs",down%image_freqs,tag)
    call to_history("firstz",down%firstz,tag)
    call to_history("firstzW",down%firstzW,tag)
    call to_history("antlialiasing",down%antion,tag)
    call to_history("num depth blocks",down%nzb)
    call to_history("num. blocks",down%nzs)
    call to_history("num freq. blocks",down%nwb)
    call to_history("num. blocks",down%nws)
    if (down%antion) then
       call to_history("kmax_x",down%pmax_x,tag)
       call to_history("kmax_y",down%pmax_y,tag)
       call to_history("kmin_x",down%pmin_x,tag)
       call to_history("kmin_y",down%pmin_y,tag)
    end if
    call to_history("eps",down%eps,tag)
    X%indx=1;X%indy=1
    i=srite(tag,0.,4)
    i=sseek(tag,0,0)

    create_image_file=T
  end function create_image_file
!--------------------------------------------------------------
!  Core routine to setup an image
  logical function image_setup(X,zero)
    type(imag), intent(inout) :: X
    integer :: er
    logical :: zero
#ifdef DEBUG
    character(len=128), parameter :: unit='image_setup'
    if(run%debug) call in(unit)
#endif

    image_setup=F
    X%bites=X%esize*&
            product((/down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,down%nzs/))

    !! DEPTH blocking
    if (.not.down%autonzs) then
       if( (1.0*X%az%n)/down%nzs /= X%az%n/down%nzs ) then
          call image_report(X)
          call seperr('R.az.n is not a multiple of nzs')
       end if
    end if
    er=0
    down%nzb = down%az%n / down%nzs
    if (.not. down%image_freqs) then
       allocate(X%bin(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,down%nzs),stat=er)
       if(er/=0) call seperr('cant allocate image sized binary',X%tag)
    else
       X%bites=X%esize*&
               product((/down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,down%nws/))
        allocate(X%bin(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,1),stat=er)
    end if

    if (zero) call image_assign(X,cmplx(0.,0.))

#ifdef SEP_MPI
    if(run%mpi) then
      if(.not. mpi_image_setup(X,.not.down%add)) then
        write(0,*)  "trouble setting up mpi image files"
        return
      end if 
    end if
#endif
    image_setup=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function image_setup
!--------------------------------------------------------------
!  Assign value to the image file
  subroutine image_assign(X,v)
    type(imag), intent(inout) :: X
    complex, intent(in) :: v
    integer :: izb,n,i,ierr
    real:: u
#ifdef DEBUG
    character(len=128), parameter :: unit='image_assign'
    if(run%debug) call in(unit)
#endif
    u=real(v)

    X%bin=v
    ierr=sseek(X%tag,0,0)
    if (X%esize==4) then
       tmpr=u
       do i=1,down%az%n
          ierr=srite(X%tag,tmpr,size(tmpr)*4)
       end do
    else
       ierr=srite(X%tag,X%bin,size(X%bin)*8)
    end if

    ierr=sseek(X%tag,0,0)

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_assign
!----------------------------------------------------------------
! Write info about image
  subroutine image_report(X)
    type(imag), intent(in) :: X
    write(0,*)'        ************************************************'
    write(0,*)'        *   ',trim(X%tag),' from ',trim(X%from)
    write(0,*)'        *                 [n]      [o]      [d]'
    write(0,'(a,i4,f10.2,f10.3)       ') "         *             x: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *             y: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *     Wrk areax: ",down%amx%n, down%amx%o, down%amx%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *     Wrk areay: ",down%amy%n, down%amy%o, down%amy%d
!    write(0,'(a,i4,f10.2,f10.3)       ') "         *          Shtx: ",down%asx%n, down%asx%o, down%asx%d
!    write(0,'(a,i4,f10.2,f10.3)       ') "         *          Shty: ",down%asy%n, down%asy%o, down%asy%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *            hx: ",X%ahx%n, X%ahx%o, X%ahx%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *            hy: ",X%ahy%n, X%ahy%o, X%ahy%d
    write(0,'(a,i4,f10.2,f10.3,a,i3,a)') "         *             z: ",X%az%n,  X%az%o,  X%az%d," ->",down%nzb," blocks"
    if(down%image_freqs)&
    write(0,'(a,i4,f10.2,f10.3)       ') "         *         freqs: ",X%aw%n, X%aw%o, X%aw%d
    if(associated(X%bin)) then
    write(0,'(a,f9.2,a,f9.2,a)        ') "         *       block: ",X%bites/1024./1024.," / ",X%esize*size(X%bin)/1024./1024.," Mbytes"
    write(0,'(a,i1)')                    "         *   esize= ",X%esize
    end if

  end subroutine image_report
!----------------------------------------------------------------
! Places single-shot image subspace into 
!          total image space. By brad 11/02/03
  subroutine image_location(X)
    type(imag), intent(inout)   ::X
    integer:: ierr,i,j,k,l,m,n,p,o
#ifdef DEBUG
    character(len=128), parameter :: unit='image_location'
    if(run%debug) call in(unit)
#endif
    X%indx= (1.5+ (down%amx%o-down%aix%o)/down%amx%d)
    X%indx= min(X%indx,down%aix%n-down%amx%n) !! Dont go off min/max points
    X%indx= max(X%indx,1)

    X%indy= (1.5+ (down%amy%o-down%aiy%o)/down%amy%d)
    X%indy= min(X%indy,down%aiy%n-down%amy%n) !! Dont go off min/max points
    X%indy= max(X%indy,1)

    X%finx=X%indx + down%amx%n -1
    X%finy=X%indy + down%amy%n -1
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_location
!----------------------------------------------------------------
! Function to read image
  subroutine image_read(X,k)
    type(imag), intent(in) :: X
    integer, intent(in) :: k
    integer:: i1,i2,i3,i4,i5,ierr,n_dim_cube,n
    integer,dimension(5)::n_cube,n_wind,f_wind,j_wind
#ifdef DEBUG
    character(len=128), parameter :: unit='image_read'
    if(run%debug) call in(unit)
#endif
    n=down%nzs
    if(down%firstz.gt.1) n=X%az%n

    n_dim_cube=5
    n_cube=(/   X%amx%n,   X%amy%n,X%ahx%n,X%ahy%n,X%az%n /)
    j_wind=(/1         ,1         ,1      ,1      ,1      /)

    if(X%esize==4) then
    n_wind=(/down%amx%n,down%amy%n,X%ahx%n,X%ahy%n,1      /)
write(0,*) 'check R:',n
write(0,*) '     xin:',X%indx -1
write(0,*) '     yin:',X%indy -1

       do i5=1,n
write(0,*) '       ',i5,':',i5-1+(k-1)*n
          f_wind=(/X%indx -1,X%indy -1,0,0,i5-1+(k-1)*n /)
          ierr=sreed_window(X%tag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,X%esize,tmpr(1,1,1,1))
          if (ierr.lt.0) then
             write(0,*) 'sreed_window return value=',ierr
             call seperr("Baddness in image_read")
          end if
          do i4=1,size(tmpr,4)
          do i3=1,size(tmpr,3)
          do i2=1,size(tmpr,2)
          do i1=1,size(tmpr,1)
             X%bin(i1,i2,i3,i4,i5)=cmplx(tmpr(i1,i2,i3,i4),0.)
          end do; end do; end do; end do
       end do
    else
       n_wind=(/down%amx%n,down%amy%n,X%ahx%n,X%ahy%n,down%nzs/)
       f_wind=(/X%indx -1,X%indy -1, 0, 0,(k-1)*n  /)
       ierr=sreed_window(X%tag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,X%esize,X%bin(1,1,1,1,1))
       if (ierr.lt.0) then
          write(0,*) 'sreed_window return value=',ierr
          call seperr("Baddness in image_read")
       end if
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_read

!----------------------------------------------------------------
!  Write out image
  subroutine image_write(X,k)
    type(imag), intent(in) :: X
    integer, intent(in) :: k
    integer:: i1,i2,i3,i4,i5,ierr,n_dim_cube,n
    integer,dimension(5)::n_cube,n_wind,f_wind,j_wind

#ifdef DEBUG
    character(len=128), parameter :: unit='image_write'
    if(run%debug) call in(unit)
#endif
    if(down%image_freqs) return
    n=down%nzs
    if(down%firstz.gt.1) n=X%az%n
    n_dim_cube=5
    n_cube=(/   X%amx%n,   X%amy%n,X%ahx%n,X%ahy%n,X%az%n /)
    j_wind=(/1         ,1         ,1      ,1      ,1        /)

    if (X%esize==4)then
    n_wind=(/down%amx%n,down%amy%n,X%ahx%n,X%ahy%n,1        /)
       do i5=1,n
          do i4=1,size(tmpr,4)
          do i3=1,size(tmpr,3)
          do i2=1,size(tmpr,2)
          do i1=1,size(tmpr,1)
             tmpr(i1,i2,i3,i4)=real(X%bin(i1,i2,i3,i4,i5))
          end do; end do ;end do ;end do
          f_wind=(/X%indx -1, X%indy -1, 0, 0, i5-1+(k-1)*n /)
          ierr=srite_window(X%tag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,X%esize,tmpr(1,1,1,1))
          if (ierr.lt.0) then
             write(0,*) 'srite_window return value=',ierr
             call seperr("Baddness in image_write")
          end if
       end do
    else
          n_wind=(/down%amx%n,down%amy%n,X%ahx%n,X%ahy%n,down%nzs /)
          f_wind=(/X%indx -1,X%indy -1,0,0,(k-1)*n  /)
          ierr=srite_window(X%tag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,X%esize,X%bin(1,1,1,1,1))
          if (ierr.lt.0) then
             write(0,*) 'srite_window return value=',ierr
             call seperr("Baddness in image_write")
          end if
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_write
!----------------------------------------------------------------
! Image initializing routine when the image file exists
  logical function image_history_init(X)
    type(imag), intent(inout) :: X

#ifdef DEBUG
    character(len=128), parameter :: unit='image_history_init'
    if(run%debug) call in(unit)
#endif
    image_history_init=F

    call auxinout(X%tag)
    X%from = "history"

    call from_aux(X%tag,"esize",X%esize,4)
    call hinaxis(X%tag,X%amx,1); X%amx%l='ix'
    call hinaxis(X%tag,X%amy,2); X%amy%l='iy'
    call hinaxis(X%tag,X%az ,5); X%az%l ='z_'
    call hinaxis(X%tag,X%ahx,3); X%ahx%l='hx'
    call hinaxis(X%tag,X%ahy,4); X%ahy%l='hy'
    down%aix=X%amx
    down%aiy=X%amy
    if(.not. image_setup(X,.false.)) then
       write(0,*) "trouble setting up image"
       return
    end if
    call image_report(X)
    image_history_init=T

#ifdef DEBUG
    if(run%debug) call image_report(X)
    if(run%debug) call out(unit)
#endif
  end function image_history_init
!----------------------------------------------------------------
  subroutine update_index(X,Y)
    type(imag),intent(in)::Y
    type(imag),intent(inout)::X

    X%indx=Y%indx; X%indy=Y%indy; X%finx=Y%finx; X%finy=Y%finy;

  end subroutine update_index
!----------------------------------------------------------------
!   Setup image structure for local files if
!     we are running MPI
  logical function setup_lead_image_file(X,lead)
    type(imag), intent(inout) :: X
    logical :: lead
    integer :: ierr,nsz
    character(len=128) :: first_lock_file,aa

#ifdef SEP_MPI
      setup_lead_image_file=.false.

      !MAKE SURE THE FIRST THREAD IS A LEAD
      call MPI_Barrier(MPI_COMM_WORLD,ierr)
      nsz=len(trim(down%image_lock))
      call MPI_Bcast(nsz,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
      first_lock_file=down%image_lock
      call MPI_Bcast(first_lock_file,nsz,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)

      !FIGURE OUT WHAT THE LEAD TRACE FOR EACH MACHINE
      down%image_lock_fd=-1
      if(first_lock_file(1:nsz)/=down%image_lock(1:nsz) .or. &
        run%m_ith == 0) &
         down%image_lock_fd=mpi_sep_check_lock(down%image_lock)
      lead=.true. 
      if(-1==down%image_lock_fd) lead=.false.

      !CREATE THE LEAD FILE
      if(lead) then
        if(.not. create_image_file(X,X%tag)) then
          write(0,*) "trouble setting up mpi data"
          return
        end if
      end if
      !WAIT UNTIL THE LEAD FILES HAVE BEEN CREATED
      call MPI_Barrier(MPI_COMM_WORLD,ierr)
      !IF WE ARENT THE LEAD FILE OPEN UP THE IMAGE FILE WITH THE RIGHT DATAPATH
      if(.not.lead) then
           call auxinout(X%tag)
!           call auxclose(X%tag)
       end if
#endif
      setup_lead_image_file=.true.
    end function setup_lead_image_file
!----------------------------------------------------------------
  logical function mpi_image_setup(X,zero)
    type(imag), intent(inout) :: X
    logical :: zero,create
    integer :: ierr,nsz,n,i
#ifdef SEP_MPI
    mpi_image_setup=F

    X%tag_0=X%tag
    if(run%share_image) then
       write(X%tag,"(a,a)") trim(X%tag_0),run%mach
    else
!       write(X%tag,"(a,a)") trim(X%tag_0),run%tag_num
    write(X%tag,*) trim(X%tag_0),trim(run%tag_num)
    i=1
    do while(i<128 .and. X%tag(i:i)==" ")
      i=i+1
    end do
    X%tag=X%tag(i:len(trim(X%tag)))
    end if 

    if(run%share_image) then
      if(.not.setup_lead_image_file(X,create)) then
        write(0,*) "trouble setting up lead image file"
        return
      end if
    else
!       X%aw_local%n=down%nw_local*down%nws

       n=X%aw%n
       X%aw%n=down%nw_local*down%nws

       create=.true.
      if(.not. create_image_file(X,X%tag)) then
        write(0,*) "trouble setting up mpi data"
        return
      end if
      X%aw%n=n
    end if 
    if(run%share_image) image_comm_group= create_group(create)
    image_lead_thread=create
    if(create) then
      if(.not. zero) then
        call distribute_image(X,create)
      else 
        call  image_assign(X,cmplx(0.,0.))
      end if
    end if
    if(create)call mpi_sep_close_lock(down%image_lock,down%image_lock_fd)
#endif
    mpi_image_setup=T
  end function mpi_image_setup
  !----------------------------------------------------------------
  !
  !  Function to combine image
  subroutine combine_image(X)
    type(imag), intent(inout) :: X
    integer ::  ierr,iadd,i
    integer, external :: sseek
    
#ifdef SEP_MPI
    call MPI_Barrier(MPI_COMM_WORLD,ierr)
    iadd=0;if(down%add) iadd=1;
    if(.not.down%image_freqs)then
       if(run%mpi .and. down%adj .and. image_lead_thread) then
          ierr=sseek(trim(X%tag),0,0) 
          if(run%m_ith==0) ierr= sseek(trim(x%tag_0),0,0)
          if(0/=mpi_sep_tag_sum(run%m_ith,X%tag,0,X%tag_0,X%bites/4,&
          down%nzb,iadd)) call seperr("trouble combining image ");
       end if
       call MPI_Barrier(MPI_COMM_WORLD,ierr)
       call aux_unlink(X%tag)
       if(var%use_disk_buf) call aux_unlink(X%tag_var)
    else
                        ierr=sseek(trim(X%tag)  ,0,0)
       if(run%m_ith==0) ierr=sseek(trim(x%tag_0),0,0)
       do i=1,X%az%n
          if(0/=mpi_sep_tag_combine(run%m_ith,X%tag,0,X%tag_0, &
          X%bites/4,down%nwp,down%iw_thread,iadd)) &
          call seperr("trouble combining image")
       end do
       call aux_unlink(X%tag)
    end if
#endif
  end subroutine combine_image

  !----------------------------------------------------------------
  !
  !  Distribute the image to various threads
  subroutine distribute_image(X,use)
    type(imag), intent(inout) :: X
    integer :: ierr,iuse,i
    logical :: use

#ifdef SEP_MPI
    if(run%mpi) then
        ierr= sseek(trim(X%tag),0,0)
       if(run%m_ith==0) then
          ierr=sseek(trim(X%tag_0),0,0)
        end if
       if(run%share_image) then
          i=1; do while(image_master(i)==0);i=i+1;end do;i=i-1;
         if(0/=mpi_sep_tag_bcast_index(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
           down%nzb,image_comm_group))call seperr("trouble distributing image 1")
       else
         if(0/=mpi_sep_tag_bcast(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
           down%nzb)) call seperr("trouble distributing image 2")
       end if
    end if
#endif
  end subroutine distribute_image
!------------------------------------------------------------------------
  integer function create_group(use)
   logical ::use
   integer :: iuse,i
   integer :: ierr
#ifdef SEP_MPI
   integer :: status(MPI_STATUS_SIZE)
#endif
   if(use) then
     iuse=1
   else 
     iuse=0
   end if

#ifdef SEP_MPI
   allocate(image_master(run%m_nth))
   do i=1,run%m_nth 
     if(run%m_ith==0 .and. i/=1) then
        call mpi_recv(image_master(i),1,MPI_INTEGER,i-1,3242+i,MPI_COMM_WORLD,status,ierr);
     else if(i-1 == run%m_ith .and. run%m_ith/=0) then
        call mpi_send(iuse,1,MPI_INTEGER,0,3242+i,MPI_COMM_WORLD,ierr)
     end if
   end do
   if(run%m_ith==0) image_master(1)=iuse
!   call MPI_Barrier(MPI_COMM_WORLD)
   call  mpi_bcast(image_master,size(image_master),MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
   create_group= mpi_sep_create_group(size(image_master),image_master)
#else
   create_group=1
#endif
 end function create_group
  !----------------------------------------------------------------
end module wei_R

