module wei_SPmig
  !! WEI migration, Shot Profile style
  use wei_util
  use wei_run
  use wei_type

  use wei_S
  use wei_R
  use wei_D
  use wei_W

  use wei_fft
  use wei_kmap
  use wei_xmap

  use wei_kxmig
  use wei_image
  use triangle2

  implicit none

  type(dat),public :: D,D_d
  type(imag),public :: R,M,N_u,N_d

  integer:: time_array(8)
  integer:: present

contains
  subroutine SPmig_init(SLin,WCin,FKin,FXin,IGin,SMin)
    integer :: st,er

#ifdef DEBUG
    character(len=128), parameter :: unit='mig_init'
#endif
    interface !!-------------------------
       integer function    SLin() result(st)
       end function SLin
       integer function    WCin() result(st)
       end function WCin
       integer function    FKin() result(st)
       end function FKin
       integer function    FXin() result(st)
       end function FXin
       integer function    IGin() result(st)
       end function IGin
       integer function    SMin() result(st)
       end function SMin
    end interface !!---------------------

    !! initialize the main WEI variables
    if(.not. run_init()) &
    call erexit("trouble initializing run_init")

    if(.not. down_init()) &
    call erexit("trouble initializing down_init")

    if(.not. init_distrib_pattern()) & ! w/o mpi, I should redo this step
    call erexit("trouble initializing distribution pattern")


#ifdef DEBUG
    if(run%debug) call in(unit)
#endif

    call file_io()

    !! convert frequency to radians
    call process_radians()
    !! init FFT
    call fft_init_r(D%bin)  
    call fft_init_s(D_d%bin)  

    !!linear maps of x and k values
    call kmap_init()
    call xmap_init()

    !! init imaging
    call imaging_init()

    !! allocate slowness arrays for FK/FX ops
    call slow_allocate()

    !! init operators
    st=SLin()     !! FK slownesses init
    st=WCin()     !! WCop init
    st=FKin()     !! FKop init
    st=FXin()     !! FXop init
    st=IGin()     !! Imaging condition init 
    st=SMin()     !! Smoothing op init

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine SPmig_init
  !----------------------------------------------------------------
  subroutine mig_outnull()
  !ZERO OUTPUT IF REQUESTED
#ifdef DEBUG
    character(len=128), parameter :: unit='mig_outnull'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! R<-D
       call     image_assign(R,(0.,0.))
    else             !! R->D
       call      data_assign(D ,(0.,0.))
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine mig_outnull
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  integer function SPmig(adj,add,mod_,dat_,dat_d_,SLop,WCop,FKop,FXop,IGop,MIGop,NDIGop,NUIGop,SMop)
  logical,            intent(in) :: adj,add
  character(len=128), intent(in) :: mod_,dat_,dat_d_
    integer,external,optional :: WCop,FKop,FXop,IGop,MIGop,NDIGop,NUIGop,SMop
    integer :: izb,iwb,isx,isy
    integer :: st

#ifdef DEBUG
    character(len=128), parameter :: unit='SPmig'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif

    !! reset tag variables
    D%tag     = trim(down%dat_)
    D_d%tag   = trim(down%dat_d_)   
    if(down%image) R%tag  = trim(down%mod_)
    if(down%mult ) M%tag  = trim(down%m_)
    if(down%normd) N_d%tag= trim(down%n_d_)
    if(down%normu) N_u%tag= trim(down%n_u_)

    !! set the adjoint sign
    if(down%adj) then
       down%opsign = -1
    else
       down%opsign = +1
    end if
    if(run%verb)write(0,*) '        * '
    if(run%verb)write(0,*) '        * Start parallel'
    call sep_begin_prog()
    if(run%verb)write(0,*) '        * '
    down%shotnum=0
    shoty: do isy=1,D%asy%n
    shotx: do isx=1,D%asx%n
       if (run%verb)then
          call line()
          write(0,'(a,i5,a,i5)')"         *  Shotnum",&
          down%shotnum+1,' of',D%asy%n*D%asx%n
       end if
       if(down%marine) then
          call data_position(D,isx,isy) 
          call xmap_op()
       else
          call data_indices_land(D_d,isx,isy)
       end if
       call data_position_report()
       if( .not.down%passive) call source_report(D_d)

       SPmig=SPmigof(isx,isy,SLop,WCop,FKop,FXop,IGop,MIGop,NDIGop,NUIGop,SMop)

       if(run%verb)write(0,'(a,i4)') '         *       Finished migrating shot',down%shotnum +1

       down%shotnum=down%shotnum+1
			 if (.not. down%teleseismic) &
  		 st=sseek(D_d%tag,0,0)               !! rewind to beginning of source
    end do shotx
    end do shoty
    call aux_unlink('distrib_pattern')
    call message("Done and Done")
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    SPmig=OK
  end function SPmig
!--------------------------------------------------------------------
  integer function SPmigof(isx,isy,SLop,WCop,FKop,FXop,IGop,MIGop,NDIGop,NUIGop,SMop)
    integer,external,optional :: WCop,FKop,FXop,IGop,MIGop,NDIGop,NUIGop,SMop
    integer :: izb,iwb,position
    integer :: st,ierr,isx,isy
#ifdef DEBUG
    character(len=128), parameter :: unit='migof'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
    var%dzb=down%az%d

    if(down%adj) then !! MIGRATION
#ifdef DEBUG
       if(run%debug) call in('migration')
#endif 
       wbA: do iwb=1,down%nwb
          if(run%verb)write(0,'(a,i2,a,i2)') &
             '         *   Frequency block ',iwb,' of ',down%nwb
          position=down%shotnum*down%nwb + (iwb-1)
          if(.not. data_read(D,position)) &
             call seperr("trouble reading D data") !! read data at block top
          if (.not.down%passive ) then !!D_d already pointed to D
             if(.not. data_read(D_d,iwb-1)) & !!read the source file
                call seperr("trouble reading D_d data")
             call data_conjg(D_d)
          else
             call data_copy_conjg(D,D_d)
          end if

          zbA: do izb=1,down%nzb
             call read_image(izb)

             call setup_var_block(iwb,izb)
             call slowness_read(S,izb) !! read a slowness block
             st = SLop() !! setup sS, sR, s(z) for working space
             down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d

             call wemig  (izb,iwb,WCop,FKop,FXop,IGop,MIGop,NDIGop,NUIGop,SMop)
             call write_image(izb)

          end do zbA
       end do wbA

#ifdef DEBUG
       if(run%debug) call out('migration')
#endif
    else         !! MODELING
#ifdef DEBUG
       if(run%debug) call in('modeling')
#endif
       !! no modeling operator yet
#ifdef DEBUG
       if(run%debug) call out('modeling')
#endif
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    SPmigof=OK
  end function SPmigof 
!---------------------------------------------------------------
  !! dont write the wavefield
  subroutine wemig(izb,iwb,WCop,FKop,FXop,IGop,MIGOP,NDIGop,NUIGop,SMop)
    integer, intent(in)       :: izb,iwb
    integer,external :: WCop,FKop,FXop,IGop,MIGOP,NDIGop,NUIGop,SMop
    integer :: izs,iws,nzs,a,currentz
    real        :: startIC,finishIC,startEX,finishEX
#ifdef DEBUG
    character(len=128), parameter :: unit='wemig'
    if(run%debug) call in(unit)
#endif
    if(down%adj) then !! MIGRATION
       if (down%image_freqs) a=sseek(R%tag,0,0)
          finishIC=0.
          finishEX=0.
       zsA: do izs=1,down%nzs          !LOOP OVER z STEPS
#ifdef TIME
          call start_timer(startIC)
#endif
          call writewavefields(izs,izb,iwb)
          call imagingcondition(izs,IGop,MIGOP,NDIGop,NUIGop,SMop)
#ifdef TIME
          call end_timer(finishIC,startIC)
          call start_timer(startEX)
#endif 

!          call kxmig(izs,D%bin,D_d%bin,WCop,FKop,FXop)

#ifdef TIME
          call end_timer(finishEX,startEX)
#endif
          currentz=(izb-1)*down%nzs+izs
       if (mod(currentz,down%az%n/10)==0) call depth_report(izs,izb,iwb)
       end do zsA
#ifdef TIME
       write(0,'(8x,a,f10.2,a)')' *       Time imaging       :', &
         finishIC,' (s)'
       write(0,'(8x,a,f10.2,a)')' *       Time extrapolating :', &
         finishEX,' (s)'
#endif
    else         !! MODELLING
       call erexit('Shot profile modeling not implemented.')
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wemig
!----------------------------------------------------------------
  subroutine writewavefields(izs,izb,iwb)
    integer::izs,izb,iwb

    !! write wavefields if you want them
    if (down%wIOu)then
       if(izs.ge.down%firstzW) then !! limit W file size when min. target depth known
          if (.not. wavefield_write(W  ,D%bin  ,iwb,izb,izs-down%firstzW+1)) &
          call seperr('wavefield_write(W ==badness') 
       end if
    end if
    if (down%wIOd)then
       if(izs.ge.down%firstzW) then !! limit W file size when min. target depth known
          if (.not. wavefield_write(W_d,D_d%bin,iwb,izb,izs-down%firstzW+1)) &
          call seperr('wavefield_write(W_d ==badness')
       end if
    endif
  end subroutine writewavefields
!----------------------------------------------------------------
  subroutine imagingcondition(izs,IGop,MIGOP,NDIGop,NUIGop,SMop)
    integer,external ::IGop,MIGOP,NDIGop,NUIGop,SMop
    integer::izs
    !! smoothed DD* for deconvolution
!    if(down%dcnv) call ddstar(D_d%bin,D_d%ddbin,SMop)

    !! all the various imaging conditions
    if(down%normu)call SPimg(izs,D%bin,D_d%bin,D_d%ddbin,N_u%bin,NUIGop)!UU*
    if(down%normd)call SPimg(izs,D%bin,D_d%bin,D_d%ddbin,N_d%bin,NDIGop)!DD*

    if(down%mult) call SPimg(izs,D%bin,D_d%bin,D_d%ddbin,M%bin  ,MIGop)!UU
    if(down%image)call SPimg(izs,D%bin,D_d%bin,D_d%ddbin,R%bin  ,IGop) !UD*

  end subroutine imagingcondition
!----------------------------------------------------------------
  subroutine depth_report(izs,izb,iwb)
    integer :: izs,izb,iwb
    integer :: pcent,current,total

    pcent=(100.0*(izs+down%nzs*(izb-1))/down%az%n/(real(down%nwb))) +&
        down%az%n*(iwb-1)/real(down%nwb) 

    current=izs+down%nzs*(izb-1)
    total=down%nzs*down%nzb
    if(run%verb) then
       write(0,'(a,i5,a,i5,a,i2,a)')'         *     Finished depth',&
       current,' of',total,' (',pcent,'%)'
    end if
  end subroutine depth_report
!----------------------------------------------------------------
  subroutine file_io()
    integer:: err
  
    down%WIOu=F;down%WIOd=F
    down%image=F;down%mult=F;
    down%normu=F;down%normd=F

    !! initialize objects
    if(.not. data_init(D,down%D)) &
    call erexit("trouble initializing D")
!    if (.not.down%passive .or. .not.down%teleseismic)then
    if (.not.down%passive )then
       if(.not. source_init(D_d,down%D_d)) & 
       call erexit("trouble initializing D_d")
    else
       if(.not. aux_data_init(D_d,down%D_d)) &
       call erexit("trouble initializing D_d")
    end if

    if (down%marine) call data_indices_marine(D,D_d)

    err=getch("R",'d',present)
    if (0.ne.err) then
       down%image=.true.
       if(.not. image_init(R,down%R)) &
         call erexit("trouble initializing image R")
          if(run%verb) write(0,*) "        *      Image Space     (UD*)          R=",down%image
    end if

    err=getch("M",'d',present)
    if (0.ne.err ) then
       down%mult=.true.
       if (.not.down%image) then
          if(.not. image_init(M,down%M)) &
             call erexit("trouble initializing image R")
          if(run%verb) write(0,*) "        *      Image Space SRMP (UU)          M=",down%mult
       else
          if(.not. aux_image_init(M,R,down%M))call seperr('badness at image_init(M)')
          if(run%verb) write(0,*) "        *      Image Space SRMP (UU)          M=",down%mult
       end if
    else
          if(.not.down%image) call seperr("You must have some output idiot.")
    end if

    err=getch("N_d",'d',present)
    if (0.ne.err )then
       down%normd=.true.
       if(down%image) then
          if(.not. aux_image_init(N_d,R,down%N_d))call seperr('badness at bwfld_init(N_d)')
          if(run%verb) write(0,*) "        *      Writing normalization (DD*)  N_d=",down%normd
       else
          if(down%mult) then
             if(.not. aux_image_init(N_d,M,down%N_d))call seperr('badness at bwfld_init(N_d)')
             if(run%verb) write(0,*) "        *      Writing normalization (DD*)  N_d=",down%normd
          else
             call seperr("You cant do that")
          end if
       end if
    end if

    err=getch("N_u",'d',present) 
    if (0.ne.err)then
       down%normu=.true.
       if(down%image) then
          if(.not. aux_image_init(N_u,R,down%N_u))call seperr('badness at image_init(N_u)')
          if(run%verb) write(0,*) "        *      Writing normalization (UU*)  N_u=",down%normu
       else
          if(down%mult) then
             if(.not. aux_image_init(N_u,M,down%N_u))call seperr('badness at bwfld_init(N_u)')
             if(run%verb) write(0,*) "        *      Writing normalization (UU*)  N_u=",down%normu
          else
             call seperr("You cant do that")
          end if
       end if
    end if

    if(.not. slow_init(S,down%VS,down%VG))&
         call erexit("trouble initializing slowness/velocity")
    
    if (0.ne.getch("W",'d',present) )then
       if(down%image_freqs)call seperr("piss off, you cant have everything")
       down%wIOu=.true.
       if(.not. bwfld_init(W,down%W))call seperr('badness at bwfld_init(W)')
       if(run%verb) write(0,*) "        *      Writing wavefield file W.    wIOu =",down%wIOu
    end if
    if (0.ne.getch("W_d",'d',present) ) then
       down%wIOd=.true.
       if(down%image_freqs)call seperr("piss off, you cant have everything")
       if(.not. bwfld_init(W_d,down%W_d))call seperr('badness at bwfld_init(W_d')
       if(run%verb) write(0,*) "        *      Writing wavefield file W_d.  wIOd =",down%wIOd
    end if

!! . .  Maping files to tags
    down%dat_    = down%D     !DATA IS FREQUENCY SLICES
    down%dat_d_  = down%D_d   !SOURCE DATA IS FREQUENCY SLICES
    if(down%image) down%mod_= down%R     !IMAGE SPACE (REFLECTIVITY)
    if(down%mult)  down%m_  = down%M !image space SRME model
    if(down%normu) down%n_u_= down%N_u !image space SRME model
    if(down%normd) down%n_d_= down%N_d !image space SRME model
  end subroutine file_io
!----------------------------------------------------------------
  subroutine read_image(izb)
    integer::izb
    if(down%image) then  !! migration output=UD*
       call image_location(R)
       call image_read(R,izb)       
       if(down%mult) then !! IS-SRMP output=UU
          call update_index(M,R)
          call image_read(M,izb)
       end if
       if(down%normd) then !! normalization output=DD* 
          call update_index(N_d,R)
          call image_read(N_d,izb)
       end if
       if(down%normu) then!! normalization output=UU*
          call update_index(N_u,R)
          call image_read(N_u,izb)
       end if
    else !! IS-SRMP w/o migration output
       call image_location(M)
       call image_read(M,izb)
       if(down%normd) then
          call update_index(N_d,M)
          call image_read(N_d,izb)
       end if
       if(down%normu) then
          call update_index(N_u,M)
          call image_read(N_u,izb)
       end if
    end if

  end subroutine read_image
!----------------------------------------------------------------
  subroutine write_image(izb)
    integer::izb
    if (.not. down%image_freqs) then
       if(down%image)call image_write(R  ,izb)!! migration output=UD*
       if(down%mult )call image_write(M  ,izb)!! IS-SRMP output=UU
       if(down%normd)call image_write(N_d,izb)!! normalization output=DD* 
       if(down%normu)call image_write(N_u,izb)!! normalization output=UU*
    end if
  end subroutine write_image
!----------------------------------------------------------------
  subroutine start_timer(time)
    real        ::time
    call date_and_time(values=time_array)
    time = time_array (5) * 3600 + time_array (6) * 60 &
         + time_array (7) + 0.001 * time_array (8)
  end subroutine start_timer
!----------------------------------------------------------------
  subroutine end_timer(ftime,stime)
    real::ftime,stime
    call date_and_time(values=time_array)
    ftime = ftime + (-stime+( time_array (5) * 3600 + time_array (6) * 60 &
                  + time_array (7) + 0.001 * time_array (8)))
  end subroutine end_timer
!----------------------------------------------------------------
end module wei_SPmig
