module wei_down
  use sep
  use wei_var
  use wei_run
  use wei_down_type
  integer,private,save :: idone,iblocks,nblocks,nzb,nwb
contains
  !----------------------------------------------------------------
  logical function down_init()
#ifdef DEBUG
    character(len=128), parameter :: unit='process_init'
    call in(unit)
#endif

    down_init=F
    !OPERATION SETUP
    call from_param('adj', down%adj, .true.)
    call from_param('add', down%add, .false.)


    !!Files expecting on the command line
    down%D    ='D'     !! receiver data
    down%D_d  ='D_d'   !! src data
    down%N_d  ='N_d'   !! DD*=source normalization
    down%N_u  ='N_u'   !! UU*=ray density
    down%M    ='M'     !! UU=SRME multiple model
    down%R    ='R'     !! image space
    down%W    ='W'     !! working space sized wavefield coming up
    down%W_d  ='W_d'   !! working space sized wavefield going down


    !! if S= is given as parameter, assumes you have slowness
    !HANDLE CONVERTED WAVE 
    ierr=getch("S_S","d",ierr)+getch("S_G","d",ierr)
    if(2==ierr) then
       down%VS='S_S'   !!       slowness
       down%VG='S_G'   !!       slowness
       down%cwave=.true.
       down%vel=.false.
    else if( 1==ierr) then
       write(0,*) "Must provide S_S and S_G"
       return
    else
       down%VS='S'
       down%VG='S'
       down%cwave=.false.
       down%vel=.false.
    end if
    !! if V= is given as parameter, assumes you have velocity
    ierr=getch("V_S","d",ierr)+getch("V_G","d",ierr)
    if(2==ierr) then
       down%VS='V_S'   !!       slowness
       down%VG='V_G'   !!       slowness
       down%cwave=.true.
       down%vel=.true.
    else if( 1==ierr) then
       write(0,*) "Must provide V_S and V_G"
       return
    else
       down%VS='V'
       down%VG='V'
       down%cwave=.false.
       down%vel=.true.
    end if

    !! get basic depth components from slowness/velocity file
    call hinaxis(down%VG,down%az,3);
        
    !HANDLE SHARED IMAGE FILE
    if(run%share_image) then
       down%image_lock="LOCK."//trim(run%mach)
    end if

    down%wIOd=.false.  !! flag whether I/O on wavefield.
    down%wIOu=.false.  !! flag whether I/O on wavefield.

    call from_param('eps',down%eps,100.)
    call from_param('scale',down%scale,1.)
    call from_param("passive",down%passive,.false.);
    call from_param("teleseismic",down%teleseismic,.false.);
    call from_param("marine",down%marine,.true.)
    if (down%passive) down%marine=.false.
    call from_param("image_real",down%image_real,.true.)
    call from_param('image_freqs',down%image_freqs,.false.)
    call from_param("firstz",down%firstz,1) !! first depth slice to write out
    call from_param("firstzW",down%firstzW,1)
    !!depth blocking info
    call from_param("nzs",down%nzs,down%az%n); down%nzb=0;
    call from_param("autonzs",down%autonzs,.true.)
    if(.not.down%autonzs.and.down%firstz.ne.1) call seperr("need autonzs to use firstz")

    if(down%image_freqs) down%autonzs=.true.
    if(down%autonzs) down%nzs=down%az%n
    if(down%firstz.gt.1) down%autonzs=T

    !! get basic freq. components from data file (D)
    call hinaxis(down%D,down%aw,3)
    if(down%aw%n==1) call erexit("Only one frequency?? Rediculous")
    !!frequency blocking info
    call from_param("nws",down%nws,down%aw%n/run%m_nth)
    call from_param("autonws",down%autonws,.true.)
    if(down%nws.gt.down%aw%n) down%autonws=.true.
    if(down%image_freqs) down%autonws=.true. !necessary??
    if (down%autonws) down%nws=down%aw%n/run%m_nth ;  down%nwb=0

    !! get width of source function
    if (.not.down%passive) then
       call hinaxis('D_d',down%fatx,1)
       call hinaxis('D_d',down%faty,2)
    else
       down%fatx%n=1; down%faty%n=1
    end if
    if(mod(down%aw%n,down%nws)/=0 ) then 
       if(run%verb)write(0,*) down%aw%n,"=down%aw%n down%nws=",down%nws
       if(run%verb)write(0,*)  "nws must be a factor of down%aw%n",mod(down%aw%n,down%nws)
       return
    end if

    !! ?should have odd number of offsets as zero is included
    call from_param('ahx_n',down%ahx%n,1)
    call from_param('ahy_n',down%ahy%n,1)
    if (mod(down%ahx%n,2)==0) down%ahx%n=down%ahx%n+1 
    if (mod(down%ahy%n,2)==0) down%ahy%n=down%ahy%n+1 

    call from_param("order",down%order,"ANY")

    if(.not. var_init()) then
       write(0,*) "trouble initializing variable step"
    end if

    if(down%order(1:3) == "VAR") then
    else if(down%order(1:3) == "ANY") then
       down%order="OZ";if(down%nzs > 10) down%order="OF"
    else if(down%order(1:2) == "OF" .or. down%order(1:2) == "OZ" ) then
    else 
       write(0,*) "invalid order ",trim(down%order), " must be OF,OZ,VAR, or ANY"
       return
    end if

    !! number of reference velocities
    call from_param("nref",down%nref,1)

    !! forward scattering migration
    call from_param("fwdscat",down%fwdscat,.false.)

    !! antialias limits
    call from_param("antialias",down%antion,.false.)
    call from_param("pmax_x",down%pmax_x,1000000.)
    call from_param("pmax_y",down%pmax_y,down%pmax_x)
    call from_param("pmin_x",down%pmin_x,-down%pmax_x)
    call from_param("pmin_y",down%pmin_y,-down%pmax_y)

    !! total shots in survey for parallel processing
    call from_param("nx_shots",down%nx_shots,1)
    call from_param("ox_shots",down%ox_shots,999999999.)
    call from_param("ny_shots",down%ny_shots,1)
    call from_param("oy_shots",down%oy_shots,999999999.)
!    down%pmax_y=1000.;down%pmin_y=-1000.

    !! smoothing parameters for deconvolution read in Smooth.F90
    call from_param("downconv",down%dcnv,.false.)
    call from_param("rectx",down%rectx,1)
    call from_param("rectx",down%recty,1)
    call from_param("rectx",down%rectw,1)

    if(run%verb) call down_report()
    down_init=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function down_init
!----------------------------------------------------------------
  subroutine down_reset2(adj,add,mod_,dat_,dat_d_)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_,dat_,dat_d_
!    character(len=128), intent(in),optional::dat_d
 
    down%adj  = adj
    down%add  = add
    if(run%restart) down%add=.true.
    down%mod_ = mod_
    down%dat_ = dat_
    down%dat_d_ = dat_d_
  end subroutine down_reset2
!----------------------------------------------------------------  
  subroutine down_report()
    call line()
    write(run%io_unit,*)              "        *Paramerter report"
    write(run%io_unit,'(a,i4,a,i4,a,i3)')  "         *  nzs=",down%nzs,"          nws=",down%nws,"           nref=",down%nref
    write(run%io_unit,*)              "        *  Marine:",down%marine,"       Passive:",down%passive,"      Velocity:",down%vel
    write(run%io_unit,*)              "        *  Image freqs:",down%image_freqs," Teleseismic:",down%teleseismic," Anti-alias:",down%antion
!    write(run%io_unit,*)              "        *  Normalize image:",down%norm,'         Multiple model:',down%mult
    write(run%io_unit,*)              "        *  Forward scatter:",down%fwdscat,'         Converted wave:',down%cwave
    if(     down%add)  write(run%io_unit,'(a,i4,a,i4)')"         *  firstZ=",down%firstz,'         add: T       firstzW=',down%firstzw
    if(.not.down%add)  write(run%io_unit,'(a,i4,a,i4)')"         *  firstZ=",down%firstz,'         add: F       firstzW=',down%firstzw
    if(down%antion) then
    write(run%io_unit,'(a,f5.4,a,f5.4)')"       *    kmax_x=",down%pmax_x,"kmin_x=",down%pmin_x
    write(run%io_unit,'(a,f5.4,a,f5.4)')"       *    kmax_y=",down%pmax_y,"kmin_y=",down%pmin_y
    end if
  end subroutine down_report
!----------------------------------------------------------------
  subroutine set_blocks(n_wb,n_zb,n_blocks)
   integer,optional :: n_zb,n_wb,n_blocks
   if(present(n_zb)) nzb=n_zb
   if(present(n_wb)) nwb=n_wb
   if(present(n_blocks)) nblocks=n_blocks
  end subroutine
  !----------------------------------------------------------------
  subroutine process_radians()
    down%aw%o = 2*pi * down%aw%o
    down%aw%d = 2*pi * down%aw%d
    down%bw%o = 2*pi * down%bw%o
    down%bw%d = 2*pi * down%bw%d
    down%pmax_x = 2* pi *down%pmax_x
    down%pmax_y = 2* pi *down%pmax_y
    down%pmin_x = 2* pi *down%pmin_x
    down%pmin_y = 2* pi *down%pmin_y 
  end subroutine process_radians
!----------------------------------------------------------------
end module wei_down
