!HANDLE FFTS
module wei_fft
  use wei_flags
  use wei_run
  use wei_S
  implicit none
  integer*8,pointer,private,save :: splanF(:),splanI(:)
  integer*8,pointer,private,save :: rplanF(:),rplanI(:)
  real, private,save :: in12
  integer::dims(2)
contains
  !----------------------------------------------------------------
  !INITIALIZE THE FFT
  subroutine fft_init_r(wfld)
    complex :: wfld(:,:,:)
    integer :: i
#ifdef DEBUG
    character(len=128), parameter :: unit='fft_init_r'
    if(run%depar) call in(unit)
#endif
    dims=(/size(wfld,1),size(wfld,2)/)
    in12=sqrt(1./real(product(dims)))

    allocate(rplanF(size(wfld,3)))
    allocate(rplanI(size(wfld,3)))
    do i=1,size(wfld,3)!!output writes over input
       call sfftw_plan_dft_2d(rplanF(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),FFTW_FORWARD,FFTW_PATIENT)
       call sfftw_plan_dft_2d(rplanI(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),FFTW_BACKWARD,FFTW_PATIENT)
    end do
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine fft_init_r
!----------------------------------------------------------------
  subroutine fft_init_s(wfld)
    complex :: wfld(:,:,:)
    integer :: i
#ifdef DEBUG
    character(len=128), parameter :: unit='fft_init_s'
    if(run%depar) call in(unit)
#endif

    allocate(splanF(size(wfld,3)))
    allocate(splanI(size(wfld,3)))

    do i=1,size(wfld,3)!!output writes over input
       call sfftw_plan_dft_2d(splanF(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),ffTW_FORWARD,fftW_PATIENT)
       call sfftw_plan_dft_2d(splanI(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),ffTW_BACKWARD,fftW_PATIENT)
    end do

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine fft_init_s
  !----------------------------------------------------------------
  !! forward FFT
  subroutine fft(iws,data)
    integer, intent(in) :: iws
    integer :: i,j
    complex, pointer, dimension(:,:,:) :: data
#ifdef DEBUG
    character(len=128), parameter :: unit='fft'
    if(run%depar) call in(unit)
    write(0,*) 'fft',iws,slo%source,omp_get_thread_num()
#endif

    if(slo%source) then
       call sfftw_execute(splanF(iws))
    else
       call sfftw_execute(rplanF(iws))
    end if

    do i=1,size(data,2)
    do j=1,size(data,1)
       data(j,i,iws)=data(j,i,iws)*in12
    end do
    end do

  end subroutine fft
!----------------------------------------------------------------
  !! inverse FFT
  subroutine ift(iws,data)
    integer, intent(in) :: iws
    integer :: i,j
    complex, dimension(:,:,:), pointer :: data
#ifdef DEBUG
    character(len=128), parameter :: unit='ift'
    if(run%depar) call in(unit)
    write(0,*) 'ift',iws,slo%source,omp_get_thread_num()
#endif

    if(slo%source) then
       call sfftw_execute(splanI(iws))
    else
       call sfftw_execute(rplanI(iws))
    end if
    do i=1,size(data,2)
    do j=1,size(data,1)
       data(j,i,iws)=data(j,i,iws)*in12
    end do
    end do

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine ift

end module wei_fft
