module wei_image
  !! WEI imaging condition

  use wei_down_type
  use wei_run
  use wei_SPig
  !---------------
  implicit none

  complex, dimension(:,:,:,:), pointer, private :: modl
contains
  !----------------------------------------------------------------  
  subroutine imaging_init()
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='image_init'
    if(run%debug) call in(unit)
#endif

    !! imaging wavefield, size of imaging space (x,y,hx,hy)
    allocate(modl(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n),stat=er)
    if(er/=0) call seperr('cant allocate modl for shot profile')
    
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine imaging_init
!------------------------------------------------------------------------------------------
  subroutine SPimg(izs,wfld_u,wfld_d,ddstar,imbin,oper)
    integer, intent(in) :: izs
    integer::i1,i2,i3,i4,i5
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    real, dimension(:,:,:), pointer :: ddstar
    complex, dimension(:,:,:,:,:), pointer :: imbin
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='SPimg'
#endif
    interface !!-------------------------
       integer function oper(image,wfld_u,wfld_d,ddstar) result(st)
         complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
         real, dimension(:,:,:), pointer :: ddstar
         complex, dimension(:,:,:,:), pointer :: image
       end function oper
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
    !! limit file size when min. target depth known
    if(izs.lt.down%firstz) return
    modl=cmplx(0.,0.)
    !! image all offsets for all frequencies
    st = oper(modl,wfld_u,wfld_d,ddstar)

    !! sum up frequency blocks from temp model into actual model
    if (.not.down%image_freqs) then
!$OMP PARALLEL DO 
          do i4=1,size(modl,4)
          do i3=1,size(modl,3)
          do i2=1,size(modl,2)
          do i1=1,size(modl,1)
             imbin(i1,i2,i3,i4,izs)=imbin(i1,i2,i3,i4,izs) + modl(i1,i2,i3,i4) !/down%aw%n
          end do;end do;end do;end do
!$OMP END PARALLEL DO
       end if
!       else
!          do i4=1,size(modl,4)
!          do i3=1,size(modl,3)
!          do i2=1,size(modl,2)
!          do i1=1,size(modl,1)
!          imbin(i1,i2,i3,i4,izs)= modl(i1,i2,i3,i4) !/down%aw%n !uncommented division
!          end do;end do;end do;end do
!       end if
!    end if

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine SPimg
!!----------------------------------------------------------------
!subroutine ddstar(wfld_d,ddstar,SmoothOp)
!    complex, dimension(:,:,:), pointer,intent(in) :: wfld_d
!    real, dimension(:,:,:), pointer,intent(out) :: ddstar
!    integer :: st
!#ifdef DEBUG
!    character(len=128), parameter :: unit='ddstar'
!#endif
!    interface !!-------------------------
!       integer function SmoothOp(wfld_d,ddstar) result(st)
!         complex, dimension(:,:,:), pointer :: wfld_d
!         real, dimension(:,:,:), pointer,intent(out) ::ddstar
!       end function SmoothOp
!    end interface !!---------------------
!    
!    st=SmoothOp(wfld,smoothamp)
!  end subroutine ddstar
!----------------------------------------------------------------
end module wei_image





