module wei_kxmig
  !! fk-fx migration
  !! receive wfld at (izs,iws)
  !! return  wfld downward/upward continued
  !! needs Fop, FXop

  use wei_util
  use wei_run
  use wei_external
  use wei_down
  use wei_S
  use wei_fft
  use wei_kmap
  !---------------
  implicit none

  !! temporary interpolation arrays
  complex, dimension(:,:,:), allocatable, private :: tmp1,tmp2
contains
!----------------------------------------------------------------
  subroutine kxmig(izs,wfld_r,wfld_s,WCop,FKop,FXop)
    complex, dimension(:,:,:), pointer :: wfld_s,wfld_r
    integer,external                ::    FKop,FXop
    integer,intent(in) :: izs 
    integer :: st,iws,ith
#ifdef DEBUG
    character(len=128), parameter :: unit='kxmig'
#endif
    interface !!-------------------------
       integer function WCop(iws,izs,ith,wfld,FKop,FXop) result(st)
         complex, dimension(:,:,:), pointer ::wfld
         integer, intent(in)       :: iws,izs,ith
         integer,external          :: FKop,FXop
       end function WCop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    !!D.C. SOURCE WAVEFIELD
    call sellect_slow_comp(.true.) !!selects slowness from source field
    ith=1
#ifdef SEP_OMP
!$OMP PARALLEL PRIVATE(ith)
    ith=1+omp_get_thread_num()
!$OMP DO PRIVATE(iws)
#endif
    wsS: do iws=1,down%nws             !LOOP OVER w STEPS
       st = WCop(iws,izs,ith,wfld_s,FKop,FXop)
    end do wsS
!$OMP END DO

    !!D.C. RECEIVER WAVEFIELD
    call sellect_slow_comp(.false.)!!selects slowness from recvr field
!$OMP DO PRIVATE(iws)
    wsR: do iws=1,down%nws             !LOOP OVER w STEPS
       st = WCop(iws,izs,ith,wfld_r,Fkop,Fxop)
    end do wsR
!$OMP END DO 
!$OMP END PARALLEL
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine kxmig
!----------------------------------------------------------------
!! N reference slownesses
  integer function kxmigN_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='kxmigN_init'
    if(run%debug) call in(unit)
#endif
    call init_wave_temps()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function kxmigN_init
!----------------------------------------------------------------
!! N reference slownesses SSR (WCop)
  integer function kxmigN_SSR(iws,izs,ith,wfld,FKop,FXop) result(st)
    complex, dimension(:,:,:), pointer     :: wfld
    integer, intent(in)                    :: iws,izs,ith
    integer                                :: ierr,ivel
#ifdef DEBUG
    character(len=128), parameter :: unit='kxmigN_SSR'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ivel,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ith,ivel
         complex, dimension(:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ivel,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ith,ivel
         complex, dimension(:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
    if(down%adj) then !! MIGRATION {
       !----------------------------------------------------------------
       call store_wavefield(wfld,iws,ith) !STORE THE CURRENT WAVEFIELD in tmp2
       refvels:do ivel=1,slo%nrs(izs)     !ref. vels
          call grab_part_wvfld(wfld ,ivel,iws,izs,ith) !GRAB PORTION OF THE WFLD FOR REF.V
          st   = FXop(iws,izs,ivel,ith, wfld) !! f-x operator
          call fft(iws,wfld)
          st   = FKop(iws,izs,ivel,ith, wfld) !! f-k operator
          call add_wfld_extraps(wfld ,iws,ith) !SUM OUR EFFORTS TOGETHER into tmp1
       end do refvels
       call grab_final_wfld(wfld,iws,ith) !copy summed up field from tmp1
       call ift(iws,wfld) 
    else             !! MODELLING {
       call seperr("No. No. No! Tell Alejandro to get to work.")
    end if!}
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function kxmigN_SSR
!----------------------------------------------------------------  
!! multiple fk ops
!!
!!       D(z)
!!       |
!!       v
!!       T1
!!
!!    D1 D2 D3 ...
!!
!!    S1 S2 S3 ...
!!     \ | /
!!       T2
!!       |
!!       v
!!       D(z+dz)
!---------------------------------------------
!  Initialize storage for wavefields
subroutine init_wave_temps() 
 integer ::er
   !! allocate tmp arrays for interpolation of multiple fk ops
   allocate(tmp1(down%amx%n,down%amy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp1 in kxmig.F90')   
   allocate(tmp2(down%amx%n,down%amy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp2 in kxmig.F90')

 end subroutine init_wave_temps
!----------------------------------------------------------------
subroutine store_wavefield(wfld,iws,ith)
  complex, dimension(:,:,:), pointer :: wfld
  integer,intent(in)::ith,iws
  integer :: i,j
  do j=1,size(tmp2,2)
  do i=1,size(tmp2,1)
     tmp2(i,j,ith) = wfld(i,j,iws)
  end do;end do

  do j=1,size(tmp1,2)
  do i=1,size(tmp1,1)
     tmp1(i,j,ith)=cmplx(0.,0.)
  end do;end do
end subroutine store_wavefield
!----------------------------------------------------------------
subroutine grab_part_wvfld(wfld,ifk,iws,izs,ith)
  complex, dimension(:,:,:), pointer :: wfld
  integer,intent(in)::ith,iws,ifk,izs
  integer :: i,j
  do j=1,size(tmp2,2)
  do i=1,size(tmp2,1)
     wfld(i,j,iws)=cmplx(0.,0.)
  end do;end do
  where(slo%si(:,:,izs) == ifk)
     wfld(:,:,iws) = tmp2(:,:,ith)
  end where
end subroutine grab_part_wvfld
!------------------------------------------------------
subroutine add_wfld_extraps(wfld,iws,ith)
  complex, dimension(:,:,:), pointer :: wfld
  integer :: i,j
  integer,intent(in)::ith,iws
  do j=1,size(tmp2,2)
  do i=1,size(tmp2,1)
     tmp1(i,j,ith) = tmp1(i,j,ith) +  wfld(i,j,iws)
  end do;end do
end subroutine add_wfld_extraps
!----------------------------------------------------------------
subroutine grab_final_wfld(wfld,iws,ith)
  complex, dimension(:,:,:), pointer :: wfld
  integer ::i,j
  integer,intent(in):: iws,ith
  do j=1,size(tmp2,2)
  do i=1,size(tmp2,1)
     wfld(i,j,iws) = tmp1(i,j,ith)
  end do
  end do
end subroutine grab_final_wfld
!----------------------------------------------------------------
subroutine store_part_wavefield(wfld,ifk,iws,izs,ith)
  complex, dimension(:,:,:), pointer:: wfld
  integer,intent(in)::ith,iws,ifk,izs
  where(slo%si(:,:,izs) == ifk)
     tmp1(:,:,ith) = tmp1(:,:,ith) + wfld(:,:,iws)
  end where
end subroutine store_part_wavefield
!----------------------------------------------------------------
end module wei_kxmig
