module wei_ssf
  !! Split-Step Fourier

  use wei_util
  use wei_run
  use wei_S
  !---------------
  implicit none
  real,  dimension(:,:,:), allocatable, private ::v

contains
!----------------------------------------------------------------
  integer function ssf_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='ssf_init'
    if(run%debug) call in(unit)
#endif
    allocate(v (size(S%pba,1),size(S%pba,2),run%o_nth) )
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function ssf_init
!----------------------------------------------------------------
  integer function ssf(iws,izs,ivel,ith,wfld) result(st)
    integer, intent(in) :: iws,izs,ivel,ith
    complex, dimension(:,:,:), pointer :: wfld
    real    ::  w,j,k,s0,ak
    complex :: iwdz
    integer :: i2,i1
#ifdef DEBUG
    character(len=128), parameter :: unit='ssf'
    if(run%depar) call in(unit)
#endif

    s0=slo%sz(izs,ivel)
    w = down%bw%o + (iws-1)*down%bw%d
    j =  w * down%opsign * var%dzb(izs)    
    if (slo%source) then
       do i2=1,size(v,2)
          do i1=1,size(v,1)
             v(i1,i2,ith)=real(S%pba(i1,i2,izs))
          end do; end do
       else
          do i2=1,size(v,2)
             do i1=1,size(v,1)
                v(i1,i2,ith)=aimag(S%pba(i1,i2,izs))
             end do; end do
          end if
          do i2=1,down%amy%n
             do i1=1,down%amx%n
                k=(v(i1,i2,ith)-s0)*j
                wfld(i1,i2,iws)=wfld(i1,i2,iws)*cmplx(cos(k),sin(k))
             end do
          end do
#ifdef DEBUG
          if(run%depar) call out(unit)
#endif
          st=OK
        end function ssf
  !----------------------------------------------------------------
end module wei_ssf
